# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-now       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
import cairo
from gi.repository import Gdk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from .libkml import Kml

# -------------------------------------------------------------------------
#
# osmGpsMap
#
# -------------------------------------------------------------------------

try:
    import gi
    from gi.repository import OsmGpsMap as osmgpsmap

    gi.require_version("OsmGpsMap", "1.0")
except:
    raise

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.kmllayer")


class KmlLayer(GObject.GObject, osmgpsmap.MapLayer):
    """
    This is the layer used to display kml files over the map
    * Allowed : points, paths and polygons.

    * One point : name, (latitude, longitude)
    * One path  : name, type, color, transparency,
    *             [ (latitude, longitude), (latitude, longitude), ...]
    * One polygon : name, type, color, transparency,
    *             [ (latitude, longitude), (latitude, longitude), ...]
    """

    def __init__(self):
        """
        Initialize the layer
        """
        GObject.GObject.__init__(self)
        self.paths = []
        self.polygons = []
        self.tag = ""
        self.name = ""
        self.type = ""
        self.points = []
        self.kml = None

    def clear(self):
        """
        reset the layer attributes.
        """
        self.paths = []
        self.polygons = []
        self.name = ""

    def add_kml(self, kml_file):
        """
        Add a kml file.
        The access right and validity must be verified before this method.
        """
        self.kml = Kml(kml_file)
        (paths, polygons) = self.kml.add_kml()
        if paths != []:
            self.paths.append(paths)
        if polygons != []:
            self.polygons.append(polygons)

    def do_draw(self, gpsmap, ctx):
        """
        Draw all the surfaces and paths
        """
        color1 = Gdk.color_parse("red")
        color2 = Gdk.color_parse("blue")
        for polygons in self.polygons:
            for polygon in polygons:
                (dummy_name, ptype, dummy_color, dummy_transparency, points) = polygon
                map_points = []
                for point in points:
                    conv_pt = osmgpsmap.MapPoint.new_degrees(point[0], point[1])
                    (coord_x, coord_y) = gpsmap.convert_geographic_to_screen(conv_pt)
                    map_points.append((coord_x, coord_y))
                first = True
                ctx.save()
                ctx.set_source_rgba(
                    float(color2.red / 65535.0),
                    float(color2.green / 65535.0),
                    float(color2.blue / 65535.0),
                    0.3,
                )  # transparency
                ctx.set_line_cap(cairo.LINE_CAP_ROUND)
                ctx.set_line_join(cairo.LINE_JOIN_ROUND)
                ctx.set_line_width(3)
                ctx.new_path()
                for idx_pt in range(0, len(map_points)):
                    if first:
                        first = False
                        ctx.move_to(map_points[idx_pt][0], map_points[idx_pt][1])
                    else:
                        ctx.line_to(map_points[idx_pt][0], map_points[idx_pt][1])
                ctx.close_path()
                if ptype == "Polygon":
                    ctx.stroke()
                if ptype == "OuterPolygon":
                    ctx.fill()
                if ptype == "InnerPolygon":
                    ctx.set_source_rgba(1.0, 1.0, 1.0, 0.3)
                    ctx.set_operator(cairo.OPERATOR_ADD)
                    ctx.fill()
                ctx.restore()
        for paths in self.paths:
            for path in paths:
                (dummy_name, ptype, dummy_color, dummy_transparency, points) = path
                map_points = []
                for point in points:
                    conv_pt = osmgpsmap.MapPoint.new_degrees(point[0], point[1])
                    (coord_x, coord_y) = gpsmap.convert_geographic_to_screen(conv_pt)
                    map_points.append((coord_x, coord_y))
                first = True
                ctx.save()
                ctx.set_source_rgba(
                    float(color1.red / 65535.0),
                    float(color1.green / 65535.0),
                    float(color1.blue / 65535.0),
                    0.5,
                )  # transparency
                ctx.set_line_width(5)
                ctx.set_operator(cairo.OPERATOR_ATOP)
                for idx_pt in range(0, len(map_points)):
                    if first:
                        first = False
                        ctx.move_to(map_points[idx_pt][0], map_points[idx_pt][1])
                    else:
                        ctx.line_to(map_points[idx_pt][0], map_points[idx_pt][1])
                ctx.stroke()
                ctx.restore()

    def do_render(self, gpsmap):
        """
        render the layer
        """
        dummy_map = gpsmap

    def do_busy(self):
        """
        set the layer busy
        """
        return False

    def do_button_press(self, gpsmap, gdkeventbutton):
        """
        When we press a button.
        """
        dummy_map = gpsmap
        dummy_evt = gdkeventbutton
        return False


GObject.type_register(KmlLayer)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2014-2015  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .editsecondary import EditSecondary
from ..glade import Glade
from ..widgets import MonitoredDate, MonitoredEntry
from ..dialog import ErrorDialog
from gramps.gen.errors import ValidationError
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("Place_Name_Editor_dialog", "manual")

ISO_CODES = (
    "aa",
    "ab",
    "ae",
    "af",
    "ak",
    "am",
    "an",
    "ar",
    "as",
    "av",
    "ay",
    "az",
    "ba",
    "be",
    "bg",
    "bh",
    "bi",
    "bm",
    "bn",
    "bo",
    "br",
    "bs",
    "ca",
    "ce",
    "ch",
    "co",
    "cr",
    "cs",
    "cu",
    "cv",
    "cy",
    "da",
    "de",
    "dv",
    "dz",
    "ee",
    "el",
    "en",
    "eo",
    "es",
    "et",
    "eu",
    "fa",
    "ff",
    "fi",
    "fj",
    "fo",
    "fr",
    "fy",
    "ga",
    "gd",
    "gl",
    "gn",
    "gu",
    "gv",
    "ha",
    "he",
    "hi",
    "ho",
    "hr",
    "ht",
    "hu",
    "hy",
    "hz",
    "ia",
    "id",
    "ie",
    "ig",
    "ii",
    "ik",
    "io",
    "is",
    "it",
    "iu",
    "ja",
    "jv",
    "ka",
    "kg",
    "ki",
    "kj",
    "kk",
    "kl",
    "km",
    "kn",
    "ko",
    "kr",
    "ks",
    "ku",
    "kv",
    "kw",
    "ky",
    "la",
    "lb",
    "lg",
    "li",
    "ln",
    "lo",
    "lt",
    "lu",
    "lv",
    "mg",
    "mh",
    "mi",
    "mk",
    "ml",
    "mn",
    "mr",
    "ms",
    "mt",
    "my",
    "na",
    "nb",
    "nd",
    "ne",
    "ng",
    "nl",
    "nn",
    "no",
    "nr",
    "nv",
    "ny",
    "oc",
    "oj",
    "om",
    "or",
    "os",
    "pa",
    "pi",
    "pl",
    "ps",
    "pt",
    "qu",
    "rm",
    "rn",
    "ro",
    "ru",
    "rw",
    "sa",
    "sc",
    "sd",
    "se",
    "sg",
    "si",
    "sk",
    "sl",
    "sm",
    "sn",
    "so",
    "sq",
    "sr",
    "ss",
    "st",
    "su",
    "sv",
    "sw",
    "ta",
    "te",
    "tg",
    "th",
    "ti",
    "tk",
    "tl",
    "tn",
    "to",
    "tr",
    "ts",
    "tt",
    "tw",
    "ty",
    "ug",
    "uk",
    "ur",
    "uz",
    "ve",
    "vi",
    "vo",
    "wa",
    "wo",
    "xh",
    "yi",
    "yo",
    "za",
    "zh",
    "zu",
)


# -------------------------------------------------------------------------
#
# EditPlaceName class
#
# -------------------------------------------------------------------------
class EditPlaceName(EditSecondary):
    """
    Displays a dialog that allows the user to edit a place name.
    """

    def __init__(self, dbstate, uistate, track, pname, callback):
        EditSecondary.__init__(self, dbstate, uistate, track, pname, callback)

    def _local_init(self):
        self.width_key = "interface.place-name-width"
        self.height_key = "interface.place-name-height"

        self.top = Glade()
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Place Name Editor")
        )

    def _setup_fields(self):
        self.value = MonitoredEntry(
            self.top.get_object("value"),
            self.obj.set_value,
            self.obj.get_value,
            self.db.readonly,
        )

        self.date = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_stat"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.language = MonitoredEntry(
            self.top.get_object("language"),
            self.obj.set_language,
            self.obj.get_language,
            self.db.readonly,
        )
        self.language.connect("validate", self._validate_iso_code)
        # force validation now with initial entry
        self.top.get_object("language").validate(force=True)

    def _validate_iso_code(self, widget, text):
        if text not in ISO_CODES:
            return ValidationError(_("Invalid ISO code"))

    def _connect_signals(self):
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_ok_button(self.top.get_object("ok"), self.save)

    def build_menu_names(self, obj):
        return (_("Place Name"), _("Place Name Editor"))

    def save(self, *obj):
        if not self.obj.get_value():
            ErrorDialog(
                _("Cannot save place name"),
                _("The place name cannot be empty"),
                parent=self.window,
            )
            return

        if self.callback:
            self.callback(self.obj)
        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham, A. Roitman
# Copyright (C) 2007-2009  B. Malengier
# Copyright (C) 2008       Lukasz Rymarczyk
# Copyright (C) 2008       Raphael Ackermann
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012       Doug Blank
# Copyright (C) 2012-2013  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Module responsible for handling the command line arguments for Gramps.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import sys
import re

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.recentfiles import recent_files
from gramps.gen.utils.file import rm_tempdir, get_empty_tempdir
from .clidbman import CLIDbManager, NAME_FILE, find_locker_name
from gramps.gen.db.utils import make_database
from gramps.gen.db.dbconst import DBBACKEND
from gramps.gen.plug import BasePluginManager
from gramps.gen.plug.report import CATEGORY_BOOK, CATEGORY_CODE, BookList
from .plug import cl_report, cl_book
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.config import config


# -------------------------------------------------------------------------
#
# private functions
#
# -------------------------------------------------------------------------
def _split_options(options_str):
    """
    Split the options for the action.

    Rules:

    * Entries in the list of options are separated by commas without
      spaces between entries
    * List values must be inclosed in brackets ("[" and "]")
    * Entries within a list value are separated by commas
    * Text values (as a value or as entries in a list) do not have to be
      enclosed in quotes unless they include commas or quotation marks.
    * Text containing double quotes must be contained in single quotes
    * Text containing single quotes must be contained in double quotes
    * Text cannot include both single and double quotes

    Examples:

    Multiple options specified::

        report -p 'name=ancestor_chart,father_disp=["$n born $b"]'

    Using text with commas and quotes::

        title="This is some text with ,s and 's"
        title='This is some text with ,s and "s'

    Using a list of text::

        textlist=[row1,row2,"row3 with ' and ,"]
    """
    name = ""
    value = ""
    parsing_value = False
    in_quotes = False
    in_list = False
    quote_type = ""
    options_str_dict = {}

    for char in options_str:
        if not parsing_value:
            # Parsing the name of the option
            if char == "=":
                # print char, "This value ends the name"
                parsing_value = True
            else:
                # print char, "This value is part of the name"
                name += char
        else:
            # Parsing the value of the option
            if value == "" and char == "[":
                # print char, "This character begins a list"
                in_list = True
                value += char
            elif in_list == True and char == "]":
                # print char, "This character ends the list"
                in_list = False
                value += char
            elif not in_quotes and (char == '"' or char == "'"):
                # print char, "This character starts a quoted string"
                in_quotes = True
                quote_type = char
                value += char
            elif in_quotes and char == quote_type:
                # print char, "This character ends a quoted string"
                in_quotes = False
                value += char
            elif not in_quotes and not in_list and char == ",":
                # print char, "This character ends the value of the option"
                options_str_dict[name] = value
                name = ""
                value = ""
                parsing_value = False
                in_quotes = False
                in_list = False
            else:
                # print char, "This character is part of the value"
                value += char

    if parsing_value and not in_quotes and not in_list:
        # Add the last option
        options_str_dict[name] = value

    return options_str_dict


# -------------------------------------------------------------------------
# ArgHandler
# -------------------------------------------------------------------------
class ArgHandler:
    """
    This class is responsible for the non GUI handling of commands.
    The handler is passed a parser object, sanitizes it, and can execute the
    actions requested working on a :class:`.DbState`.
    """

    def __init__(self, dbstate, parser, sessionmanager, errorfunc=None, gui=False):
        self.dbstate = dbstate
        self.smgr = sessionmanager
        self.errorfunc = errorfunc
        self.gui = gui
        self.user = sessionmanager.user
        if self.gui:
            self.actions = []
            self.list = False
            self.list_more = False
            self.open_gui = None
        else:
            self.actions = parser.actions
            self.list = parser.list
            self.list_more = parser.list_more
            self.list_table = parser.list_table
            self.database_names = parser.database_names
        self.open_gui = parser.open_gui
        self.imp_db_path = None
        self.dbman = CLIDbManager(self.dbstate)
        self.force_unlock = parser.force_unlock
        self.cl_bool = False
        self.imports = []
        self.exports = []
        self.removes = parser.removes
        self.username = parser.username
        self.password = parser.password

        self.open = self.__handle_open_option(parser.open, parser.create)
        self.sanitize_args(parser.imports, parser.exports)

    def __error(self, msg1, msg2=None):
        """
        Output an error. Uses errorfunc if given, otherwise a simple print.
        """
        if self.errorfunc:
            self.errorfunc(msg1)
        else:
            print(msg1, file=sys.stderr)
            if msg2 is not None:
                print(msg2, file=sys.stderr)

    # -------------------------------------------------------------------------
    # Argument parser: sorts out given arguments
    # -------------------------------------------------------------------------
    def sanitize_args(self, importlist, exportlist):
        """
        Check the lists with open, exports, imports, and actions options.
        """
        for value, family_tree_format in importlist:
            self.__handle_import_option(value, family_tree_format)
        for value, family_tree_format in exportlist:
            self.__handle_export_option(value, family_tree_format)

    def __handle_open_option(self, value, create):
        """
        Handle the "-O" or "--open" and "-C" or "--create" options.
        Only Family trees or a dir with a family tree can be opened.
        If create is True, then create the tree if it doesn't exist.
        """
        if value is None:
            return None
        db_path = self.__deduce_db_path(value)

        if db_path:
            # We have a potential database path.
            # Check if it is good.
            if not self.check_db(db_path, self.force_unlock):
                sys.exit(1)
            if create:
                self.__error(
                    _(
                        "Error: Family Tree '%s' already exists.\n"
                        "The '-C' option cannot be used."
                    )
                    % value
                )
                sys.exit(1)
            return db_path
        elif create:
            # create the tree here, and continue
            dbid = config.get("database.backend")
            db_path, title = self.dbman.create_new_db_cli(title=value, dbid=dbid)
            return db_path
        else:
            self.__error(
                _(
                    'Error: Input Family Tree "%s" does not exist.\n'
                    "If GEDCOM, Gramps-xml or grdb, use the -i option "
                    "to import into a Family Tree instead."
                )
                % value
            )
            sys.exit(1)

    def __handle_import_option(self, value, family_tree_format):
        """
        Handle the "-i" or "--import" option.
        Only Files supported by a plugin can be imported, so not Family Trees.
        """
        fname = value
        fullpath = os.path.abspath(os.path.expanduser(fname))
        if fname != "-" and not os.path.exists(fullpath):
            self.__error(_("Error: Import file %s not found.") % fname)
            sys.exit(1)

        if family_tree_format is None:
            # Guess the file format based on the file extension.
            # This will get the lower case extension without a period,
            # or an empty string.
            family_tree_format = os.path.splitext(fname)[-1][1:].lower()

        pmgr = BasePluginManager.get_instance()
        plugin_found = False
        for plugin in pmgr.get_import_plugins():
            if family_tree_format == plugin.get_extension():
                plugin_found = True

        if plugin_found:
            self.imports.append((fname, family_tree_format))
        else:
            self.__error(
                _(
                    'Error: Unrecognized type: "%(format)s" for '
                    "import file: %(filename)s"
                )
                % {"format": family_tree_format, "filename": fname}
            )
            sys.exit(1)

    def __handle_export_option(self, value, family_tree_format):
        """
        Handle the "-e" or "--export" option.

        .. note:: this can only happen in the CLI version.
        """
        if self.gui:
            return
        fname = value
        if fname == "-":
            fullpath = "-"
        else:
            fullpath = os.path.abspath(os.path.expanduser(fname))
            if os.path.exists(fullpath):
                message = (
                    _(
                        "WARNING: Output file already exists!\n"
                        "WARNING: It will be overwritten:\n   %s"
                    )
                    % fullpath
                )
                accepted = self.user.prompt(
                    _("OK to overwrite?"),
                    message,
                    _("yes"),
                    _("no"),
                    default_label=_("yes"),
                )
                if accepted:
                    self.__error(_("Will overwrite the existing file: %s") % fullpath)
                else:
                    sys.exit(1)

        if family_tree_format is None:
            # Guess the file format based on the file extension.
            # This will get the lower case extension without a period,
            # or an empty string.
            family_tree_format = os.path.splitext(fname)[-1][1:].lower()

        pmgr = BasePluginManager.get_instance()
        plugin_found = False
        for plugin in pmgr.get_export_plugins():
            if family_tree_format == plugin.get_extension():
                plugin_found = True

        if plugin_found:
            self.exports.append((fullpath, family_tree_format))
        else:
            self.__error(_("ERROR: Unrecognized format for export file %s") % fname)
            sys.exit(1)

    def __deduce_db_path(self, db_name_or_path):
        """
        Attempt to find a database path for the given parameter.

        :returns: The path to a Gramps DB or None if a database can not be
                  deduced.
        """
        # First, check if this is the name of a family tree
        db_path = self.dbman.get_family_tree_path(db_name_or_path)

        if db_path is None:
            # This is not a known database name.
            # Check if the user provided a db path instead.
            fullpath = os.path.abspath(os.path.expanduser(db_name_or_path))
            if os.path.isdir(fullpath):
                # The user provided a directory. Check if it is a valid tree.
                name_file_path = os.path.join(fullpath, NAME_FILE)
                if os.path.isfile(name_file_path):
                    db_path = fullpath

        return db_path

    # -------------------------------------------------------------------------
    # Overall argument handler:
    # sorts out the sequence and details of operations
    # -------------------------------------------------------------------------
    def handle_args_gui(self):
        """
        Method to handle the arguments that can be given for a GUI session.

        :returns: the filename of the family tree that should be opened if
                  user just passed a famtree or a filename.

        1. no options: a family tree can be given, if so, this name is tested
           and returned. If a filename, it is imported in a new db and name of
           new db returned
        2. an open and/or import option can have been given, if so, this is
           handled, and None is returned
        """
        if self.open_gui:
            # First check if a Gramps database was provided
            # (either a database path or a database name)
            db_path = self.__deduce_db_path(self.open_gui)

            if not db_path:
                # Apparently it is not a database. See if it is a file that
                # can be imported.
                db_path, title = self.dbman.import_new_db(self.open_gui, self.user)

            if db_path:
                # Test if not locked or problematic
                if not self.check_db(db_path, self.force_unlock):
                    sys.exit(1)
                # Add the file to the recent items
                title = self.dbstate.db.get_dbname()
                if not title:
                    title = db_path
                recent_files(db_path, title)
                self.open = db_path
                self.__open_action()
            else:
                sys.exit(_("Error: cannot open '%s'") % self.open_gui)
            return db_path

        # if not open_gui, parse any command line args. We can only have one
        #  open argument, and perhaps some import arguments
        self.__open_action()
        self.__import_action()
        return None

    def handle_args_cli(self, cleanup=True):
        """
        Depending on the given arguments, import or open data, launch
        session, write files, and/or perform actions.

        :param: climan: the manager of a CLI session
        :type: :class:`.CLIManager` object
        """
        # Handle the "-l" List Family Trees option.
        if self.list:
            print(_("List of known Family Trees in your database path\n"))

            for name, dirname in sorted(
                self.dbman.family_tree_list(), key=lambda pair: pair[0].lower()
            ):
                if self.database_names is None or any(
                    [
                        (re.match("^" + dbname + "$", name) or dbname == name)
                        for dbname in self.database_names
                    ]
                ):
                    print(
                        _('%(full_DB_path)s with name "%(f_t_name)s"')
                        % {"full_DB_path": dirname, "f_t_name": name}
                    )
            return

        # Handle the "--remove" Family Tree
        if self.removes:
            for name in self.removes:
                self.dbman.remove_database(name, self.user)
            return

        # Handle the "-L" List Family Trees in detail option.
        if self.list_more:
            self.dbman.print_family_tree_summaries(self.database_names)
            return

        # Handle the "-t" List Family Trees, tab delimited option.
        if self.list_table:
            print(_("Gramps Family Trees:"))
            summary_list = self.dbman.family_tree_summary(self.database_names)
            if not summary_list:
                return
            # We have to construct the line elements together, to avoid
            # insertion of blank spaces when print on the same line is used
            line_list = [_("Family Tree")]
            for key in sorted(summary_list[0]):
                if key != _("Family Tree"):
                    line_list += [key]
            print("\t".join(line_list))
            for summary in sorted(
                summary_list, key=lambda sum: sum[_("Family Tree")].lower()
            ):
                line_list = [(_('"%s"') % summary[_("Family Tree")])]
                for item in sorted(summary):
                    if item != _("Family Tree"):
                        # Translators: used in French+Russian, ignore otherwise
                        line_list += [(_('"%s"') % summary[item])]
                print("\t".join(line_list))
            return

        self.__open_action()
        self.__import_action()

        for action, op_string in self.actions:
            print(_("Performing action: %s.") % action, file=sys.stderr)
            if op_string:
                print(_("Using options string: %s") % op_string, file=sys.stderr)
            self.cl_action(action, op_string)

        for expt in self.exports:
            print(
                _("Exporting: file %(filename)s, format %(format)s.")
                % {"filename": expt[0], "format": expt[1]},
                file=sys.stderr,
            )
            self.cl_export(expt[0], expt[1])

        if cleanup:
            self.cleanup()

    def cleanup(self):
        """clean up any remaining files"""
        print(_("Cleaning up."), file=sys.stderr)
        # remove files in import db subdir after use
        self.dbstate.db.close()
        if self.imp_db_path:
            rm_tempdir(self.imp_db_path)

    def __import_action(self):
        """
        Take action for all given import files.

        .. note:: Family trees are not supported.

        If a family tree is open, the import happens on top of it. If not
        open, a new family tree is created, and the import done. If this
        is CLI, the created tree is deleted at the end (as some action will
        have happened that is now finished), if this is GUI, it is opened.
        """
        if self.imports:
            self.cl_bool = bool(self.exports or self.actions or self.cl_bool)

            if not self.open:
                # Create empty dir for imported database(s)
                if self.gui:
                    dbid = config.get("database.backend")
                    self.imp_db_path, title = self.dbman.create_new_db_cli(dbid=dbid)
                else:
                    self.imp_db_path = get_empty_tempdir("import_dbdir")
                    dbid = config.get("database.backend")
                    newdb = make_database(dbid)
                    versionpath = os.path.join(self.imp_db_path, str(DBBACKEND))
                    with open(versionpath, "w") as version_file:
                        version_file.write(dbid)

                try:
                    self.smgr.open_activate(
                        self.imp_db_path, self.username, self.password
                    )
                    msg = _("Created empty Family Tree successfully")
                    print(msg, file=sys.stderr)
                except:
                    print(_("Error opening the file."), file=sys.stderr)
                    print(_("Exiting..."), file=sys.stderr)
                    sys.exit(1)

            for imp in self.imports:
                msg = _("Importing: file %(filename)s, format %(format)s.") % {
                    "filename": imp[0],
                    "format": imp[1],
                }
                print(msg, file=sys.stderr)
                self.cl_import(imp[0], imp[1])

    def __open_action(self):
        """
        Take action on a family tree dir to open. It will be opened in the
        session manager
        """
        if self.open:
            # Family Tree to open was given. Open it
            # Then go on and process the rest of the command line arguments.
            self.cl_bool = bool(self.exports or self.actions)

            # we load this file for use
            try:
                self.smgr.open_activate(self.open, self.username, self.password)
                print(_("Opened successfully!"), file=sys.stderr)
            except:
                print(_("Error opening the file."), file=sys.stderr)
                print(_("Exiting..."), file=sys.stderr)
                sys.exit(1)

    def check_db(self, dbpath, force_unlock=False):
        """
        Test a given family tree path if it can be opened.
        """
        # Test if not locked or problematic
        if force_unlock:
            self.dbman.break_lock(dbpath)
        if self.dbman.is_locked(dbpath):
            self.__error(
                (_("Database is locked, cannot open it!") + "\n" + _("  Info: %s"))
                % find_locker_name(dbpath)
            )
            return False
        if self.dbman.needs_recovery(dbpath):
            self.__error(_("Database needs recovery, cannot open it!"))
            return False
        if self.dbman.backend_unavailable(dbpath):
            self.__error(_("Database backend unavailable, cannot open it!"))
            return False
        return True

    # -------------------------------------------------------------------------
    #
    # Import handler
    #
    # -------------------------------------------------------------------------
    def cl_import(self, filename, family_tree_format):
        """
        Command-line import routine.
        Try to import filename using the family_tree_format.
        """
        pmgr = BasePluginManager.get_instance()
        for plugin in pmgr.get_import_plugins():
            if family_tree_format == plugin.get_extension():
                import_function = plugin.get_import_function()
                import_function(self.dbstate.db, filename, self.user)

    # -------------------------------------------------------------------------
    #
    # Export handler
    #
    # -------------------------------------------------------------------------
    def cl_export(self, filename, family_tree_format):
        """
        Command-line export routine.
        Try to write into filename using the family_tree_format.
        """
        pmgr = BasePluginManager.get_instance()
        for plugin in pmgr.get_export_plugins():
            if family_tree_format == plugin.get_extension():
                export_function = plugin.get_export_function()
                export_function(self.dbstate.db, filename, self.user)

    # -------------------------------------------------------------------------
    #
    # Action handler
    #
    # -------------------------------------------------------------------------
    def cl_action(self, action, options_str):
        """
        Command-line action routine. Try to perform specified action.
        """
        pmgr = BasePluginManager.get_instance()
        if action == "report":
            try:
                options_str_dict = _split_options(options_str)
            except:
                options_str_dict = {}
                print(_("Ignoring invalid options string."), file=sys.stderr)

            name = options_str_dict.pop("name", None)
            _cl_list = pmgr.get_reg_reports(gui=False)
            if name:
                for pdata in _cl_list:
                    if name == pdata.id:
                        mod = pmgr.load_plugin(pdata)
                        if not mod:
                            # import of plugin failed
                            return
                        category = pdata.category
                        report_class = getattr(mod, pdata.reportclass)
                        options_class = getattr(mod, pdata.optionclass)
                        if category in (CATEGORY_BOOK, CATEGORY_CODE):
                            options_class(
                                self.dbstate.db, name, category, options_str_dict
                            )
                        else:
                            cl_report(
                                self.dbstate.db,
                                name,
                                category,
                                report_class,
                                options_class,
                                options_str_dict,
                            )
                        return
                # name exists, but is not in the list of valid report names
                msg = _("Unknown report name.")
            else:
                msg = _(
                    "Report name not given. "
                    "Please use one of %(donottranslate)s=reportname"
                ) % {"donottranslate": "[-p|--options] name"}

            print(_("%s\n Available names are:") % msg, file=sys.stderr)
            for pdata in sorted(_cl_list, key=lambda pdata: pdata.id.lower()):
                # Print cli report name ([item[0]), GUI report name (item[4])
                if len(pdata.id) <= 25:
                    print(
                        "   %s%s- %s"
                        % (pdata.id, " " * (26 - len(pdata.id)), pdata.name),
                        file=sys.stderr,
                    )
                else:
                    print("   %s\t- %s" % (pdata.id, pdata.name), file=sys.stderr)

        elif action == "tool":
            from gramps.gui.plug import tool

            try:
                options_str_dict = dict(
                    [tuple(chunk.split("=")) for chunk in options_str.split(",")]
                )
            except:
                options_str_dict = {}
                print(_("Ignoring invalid options string."), file=sys.stderr)

            name = options_str_dict.pop("name", None)
            _cli_tool_list = pmgr.get_reg_tools(gui=False)
            if name:
                for pdata in _cli_tool_list:
                    if name == pdata.id:
                        mod = pmgr.load_plugin(pdata)
                        if not mod:
                            # import of plugin failed
                            return
                        category = pdata.category
                        tool_class = getattr(mod, pdata.toolclass)
                        options_class = getattr(mod, pdata.optionclass)
                        tool.cli_tool(
                            dbstate=self.dbstate,
                            name=name,
                            category=category,
                            tool_class=tool_class,
                            options_class=options_class,
                            options_str_dict=options_str_dict,
                            user=self.user,
                        )
                        return
                msg = _("Unknown tool name.")
            else:
                msg = _(
                    "Tool name not given. "
                    "Please use one of %(donottranslate)s=toolname."
                ) % {"donottranslate": "[-p|--options] name"}

            print(_("%s\n Available names are:") % msg, file=sys.stderr)
            for pdata in sorted(_cli_tool_list, key=lambda pdata: pdata.id.lower()):
                # Print cli report name ([item[0]), GUI report name (item[4])
                if len(pdata.id) <= 25:
                    print(
                        "   %s%s- %s"
                        % (pdata.id, " " * (26 - len(pdata.id)), pdata.name),
                        file=sys.stderr,
                    )
                else:
                    print("   %s\t- %s" % (pdata.id, pdata.name), file=sys.stderr)

        elif action == "book":
            try:
                options_str_dict = _split_options(options_str)
            except:
                options_str_dict = {}
                print(_("Ignoring invalid options string."), file=sys.stderr)

            name = options_str_dict.pop("name", None)
            book_list = BookList("books.xml", self.dbstate.db)
            if name:
                if name in book_list.get_book_names():
                    cl_book(
                        self.dbstate.db,
                        name,
                        book_list.get_book(name),
                        options_str_dict,
                    )
                    return
                msg = _("Unknown book name.")
            else:
                msg = _(
                    "Book name not given. "
                    "Please use one of %(donottranslate)s=bookname."
                ) % {"donottranslate": "[-p|--options] name"}

            print(_("%s\n Available names are:") % msg, file=sys.stderr)
            for name in sorted(book_list.get_book_names()):
                print("   %s" % name, file=sys.stderr)

        else:
            print(_("Unknown action: %s.") % action, file=sys.stderr)
            sys.exit(1)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2009-2010  Gary Burton
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..views.treemodels.placemodel import PlaceTreeModel
from .baseselector import BaseSelector
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SelectPlace
#
# -------------------------------------------------------------------------
class SelectPlace(BaseSelector):
    def _local_init(self):
        """
        Perform local initialisation for this class
        """
        self.setup_configs("interface.place-sel", 600, 450)

    def get_window_title(self):
        return _("Select Place")

    def get_model_class(self):
        return PlaceTreeModel

    def get_column_titles(self):
        return [
            (_("Name"), 200, BaseSelector.TEXT, 0),
            (_("ID"), 75, BaseSelector.TEXT, 1),
            (_("Type"), 100, BaseSelector.TEXT, 3),
            (_("Title"), 300, BaseSelector.TEXT, 2),
            (_("Last Change"), 150, BaseSelector.TEXT, 9),
        ]

    def get_from_handle_func(self):
        return self.db.get_place_from_handle

    def setup_filter(self):
        """Build the default filters and add them to the filter menu.
        This overrides the baseselector method because we use the hidden
        COL_SEARCH (11) that has alt names as well as primary name for name
        searching"""
        cols = [
            (pair[3], pair[1] if pair[1] else 11, pair[0] in self.exact_search())
            for pair in self.column_order()
            if pair[0]
        ]
        self.search_bar.setup_filter(cols)

    def get_config_name(self):
        return __name__

    WIKI_HELP_PAGE = URL_MANUAL_SECT2
    WIKI_HELP_SEC = _("Select_Place_selector", "manual")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009-2010  Stephen George
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2011       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import sys
import os
import platform
import signal
import logging

LOG = logging.getLogger(".")

from subprocess import Popen, PIPE

# -------------------------------------------------------------------------
# process 'safe mode'; set up for a temp directory for user data
# actual directory paths set up in const module
if "-S" in sys.argv or "--safe" in sys.argv:
    from tempfile import TemporaryDirectory

    tempdir = TemporaryDirectory(prefix="gramps_")
    os.environ["SAFEMODE"] = tempdir.name

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .gen.const import APP_GRAMPS, USER_DIRLIST, USER_DATA, ORIG_HOME_DIR
from .gen.constfunc import mac
from .version import VERSION_TUPLE
from .gen.constfunc import win, get_env_var
from .gen.config import config
from .gen.errors import HandleError

# -------------------------------------------------------------------------
#
# Instantiate Localization
#
# -------------------------------------------------------------------------

from .gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Ensure that output is encoded correctly to stdout and
# stderr. This is much less cumbersome and error-prone than
# encoding individual outputs:
#
# -------------------------------------------------------------------------

try:
    # On Darwin sys.getdefaultencoding() is correct, on Win32 it's
    # sys.stdout.encoding, and on Linux they're both right.
    if mac():
        _encoding = sys.getdefaultencoding()
    else:
        _encoding = sys.stdout.encoding
except:
    _encoding = "UTF-8"
try:
    sys.stdout = open(
        sys.stdout.fileno(),
        mode="w",
        encoding=_encoding,
        buffering=1,
        errors="backslashreplace",
    )
    sys.stderr = open(
        sys.stderr.fileno(),
        mode="w",
        encoding=_encoding,
        buffering=1,
        errors="backslashreplace",
    )
except:
    pass

# -------------------------------------------------------------------------
#
# Setup logging
#
# Ideally, this needs to be done before any Gramps modules are
# imported, so that any code that is executed as the modules are
# imported can log errors or warnings.  const and constfunc have to be
# imported before this code is executed because they are used in this
# code. That unfortunately initializes GrampsLocale, so it has its own
# logging setup during initialization.
# -------------------------------------------------------------------------
"""Set up basic logging support."""

# Setup a formatter
form = logging.Formatter(
    fmt="%(asctime)s.%(msecs).03d: %(levelname)s: "
    "%(filename)s: line %(lineno)d: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

# Create the log handlers
if win():
    # If running in GUI mode redirect stdout and stderr to log file
    if not sys.stdout:
        logfile = os.path.join(USER_DATA, "Gramps%s%s.log") % (
            VERSION_TUPLE[0],
            VERSION_TUPLE[1],
        )
        # We now carry out the first step in build_user_paths(), to make sure
        # that the user home directory is available to store the log file. When
        # build_user_paths() is called, the call is protected by a try...except
        # block, and any failure will be logged. However, if the creation of the
        # user directory fails here, there is no way to report the failure,
        # because stdout/stderr are not available, and neither is the logfile.
        if os.path.islink(USER_DATA):
            pass  # ok
        elif not os.path.isdir(USER_DATA):
            os.makedirs(USER_DATA)
        sys.stdout = sys.stderr = open(logfile, "w", encoding="utf-8")
# macOS sets stderr to /dev/null when running without a terminal,
# e.g. if Gramps.app is lauched by double-clicking on it in
# finder. Write to a file instead.
if mac() and not sys.stdin.isatty():
    from tempfile import gettempdir

    log_file_name = "gramps-" + str(os.getpid()) + ".log"
    log_file_path = os.path.join(gettempdir(), log_file_name)
    log_file_handler = logging.FileHandler(log_file_path, mode="a", encoding="utf-8")
    log_file_handler.setFormatter(form)
    log_file_handler.setLevel(logging.DEBUG)

    logger = logging.getLogger()
    logger.setLevel(logging.WARNING)
    logger.addHandler(log_file_handler)
else:
    stderrh = logging.StreamHandler(sys.stderr)
    stderrh.setFormatter(form)
    stderrh.setLevel(logging.DEBUG)

    # Setup the base level logger, this one gets
    # everything.
    l = logging.getLogger()
    l.setLevel(logging.WARNING)
    l.addHandler(stderrh)


def exc_hook(err_type, value, t_b):
    """put a hook on to catch any completely unhandled exceptions."""
    if err_type == KeyboardInterrupt:
        # Ctrl-C is not a bug.
        return
    if err_type == IOError:
        # strange Windows logging error on close
        return
    if err_type == HandleError and "not found" in value.value:
        # tell Gramps to run check & repair on next start
        config.set("behavior.runcheck", True)
        config.save()
    # Use this to show variables in each frame:
    # from gramps.gen.utils.debug import format_exception
    import traceback

    LOG.error(
        "Unhandled exception\n"
        + "".join(traceback.format_exception(err_type, value, t_b))
    )


sys.excepthook = exc_hook

from .gen.mime import mime_type_is_defined


# -------------------------------------------------------------------------
#
# Minimum version check
#
# -------------------------------------------------------------------------

MIN_PYTHON_VERSION = (3, 8, 0, "", 0)
if not sys.version_info >= MIN_PYTHON_VERSION:
    logging.warning(
        _(
            "Your Python version does not meet the "
            "requirements. At least Python %(v1)d.%(v2)d.%(v3)d is needed to"
            " start Gramps.\n\n"
            "Gramps will terminate now."
        )
        % {
            "v1": MIN_PYTHON_VERSION[0],
            "v2": MIN_PYTHON_VERSION[1],
            "v3": MIN_PYTHON_VERSION[2],
        }
    )
    sys.exit(1)

# -------------------------------------------------------------------------
#
# Gramps libraries
#
# -------------------------------------------------------------------------
try:
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
except:
    pass

args = sys.argv


def build_user_paths():
    """check/make user-dirs on each Gramps session"""
    for path in USER_DIRLIST:
        if os.path.islink(path):
            pass  # ok
        elif not os.path.isdir(path):
            os.makedirs(path)


def show_settings():
    """
    Shows settings of all of the major components.
    """
    py_str = "%d.%d.%d" % sys.version_info[:3]
    try:
        import gi

        gi.require_version("Gtk", "3.0")
        from gi.repository import Gtk

        try:
            gtkver_str = "%d.%d.%d" % (
                Gtk.get_major_version(),
                Gtk.get_minor_version(),
                Gtk.get_micro_version(),
            )
        except:  # any failure to 'get' the version
            gtkver_str = "unknown version"
    except (ImportError, ValueError):
        gtkver_str = "not found"
    # no DISPLAY is a RuntimeError in an older pygtk (e.g. 2.17 in Fedora 14)
    except RuntimeError:
        gtkver_str = "DISPLAY not set"
    # exept TypeError: To handle back formatting on version split

    try:
        from gi.repository import GObject

        try:
            pygobjectver_str = "%d.%d.%d" % GObject.pygobject_version
        except:  # any failure to 'get' the version
            pygobjectver_str = "unknown version"

    except ImportError:
        pygobjectver_str = "not found"

    try:
        from gi.repository import Pango

        try:
            pangover_str = Pango.version_string()
        except:  # any failure to 'get' the version
            pangover_str = "unknown version"

    except ImportError:
        pangover_str = "not found"

    try:
        import cairo

        try:
            pycairover_str = "%d.%d.%d" % cairo.version_info
            cairover_str = cairo.cairo_version_string()
        except:  # any failure to 'get' the version
            pycairover_str = "unknown version"
            cairover_str = "unknown version"

    except ImportError:
        pycairover_str = "not found"
        cairover_str = "not found"

    try:
        gi.require_version("PangoCairo", "1.0")
        from gi.repository import PangoCairo

        pangocairo_str = PangoCairo._version
    except ImportError:
        pangocairo_str = _("not found")

    try:
        from gi import Repository

        repository = Repository.get_default()
        if repository.enumerate_versions("OsmGpsMap"):
            import gi

            gi.require_version("OsmGpsMap", "1.0")
            from gi.repository import OsmGpsMap as osmgpsmap

            try:
                osmgpsmap_str = osmgpsmap._version
            except:  # any failure to 'get' the version
                osmgpsmap_str = "unknown version"
        else:
            osmgpsmap_str = "not found"

    except ImportError:
        osmgpsmap_str = "not found"

    try:
        import gi

        repository = gi.Repository.get_default()
        if repository.enumerate_versions("GExiv2"):
            gi.require_version("GExiv2", "0.10")
            from gi.repository import GExiv2

            try:
                gexiv2_str = GExiv2._version
            except:  # any failure to 'get' the version
                gexiv2_str = "unknown version"
        else:
            gexiv2_str = "not found"

    except ImportError:
        gexiv2_str = "not found"
    except ValueError:
        gexiv2_str = "not new enough"

    try:
        vers_str = Popen(["exiv2", "-V"], stdout=PIPE).communicate(input=None)[0]
        try:
            if isinstance(vers_str, bytes) and sys.stdin.encoding:
                vers_str = vers_str.decode(sys.stdin.encoding)
            indx = vers_str.find("exiv2 ") + 6
            vers_str = vers_str[indx : indx + 4]
        except Exception:
            vers_str = _("not found")

    except Exception:
        vers_str = _("not found because exiv2 is not installed")

    try:
        import PyICU

        try:
            pyicu_str = PyICU.VERSION
            icu_str = PyICU.ICU_VERSION
        except:  # any failure to 'get' the version
            pyicu_str = "unknown version"
            icu_str = "unknown version"

    except ImportError:
        pyicu_str = "not found"
        icu_str = "not found"

    try:
        import icu

        icu_str = icu.PY_VERSION
    except Exception:
        icu_str = _("not found")

    try:
        import PIL

        try:
            pil_ver = PIL.__version__
        except Exception:
            try:
                pil_ver = str(PIL.PILLOW_VERSION)
            except Exception:
                try:
                    # print(dir(PIL))
                    pil_ver = str(PIL.VERSION)
                except Exception:
                    pil_ver = _("Installed but does not supply version")
    except ImportError:
        pil_ver = _("not found")

    def verstr(nums):
        return ".".join(str(num) for num in nums)

    try:
        gi.require_version("Gspell", "1")
        from gi.repository import Gspell

        gspell_ver = str(Gspell._version)
    except Exception:
        gspell_ver = _("not found")

    # RCS_MIN_VER = (5, 9, 4)
    # rcs_ver_str = verstr(RCS_MIN_VER)
    # print("os.environ: %s " % os.environ)
    try:
        if win():
            _RCS_FOUND = os.system("rcs -V >nul 2>nul") == 0
            rcs_ver = _("installed")
            # print("rcs -V : " + os.system("rcs -V"))
            if _RCS_FOUND and "TZ" not in os.environ:
                # RCS requires the "TZ" variable be set.
                os.environ["TZ"] = str(time.timezone)
        else:
            import subprocess
            import re

            # print("xx rcs -V : " + os.system("rcs --version"))
            _RCS_FOUND = str(subprocess.check_output(["rcs", "--version"]))
            version = re.compile(r".*GNU RCS\) ([0-9]+\.[0-9]+\.[0-9]+).*")
            rcs_ver = version.findall(_RCS_FOUND)[0]
    except Exception:
        rcs_ver = _("not found")

    try:
        gi.require_version("GeocodeGlib", "1.0")
        from gi.repository import GeocodeGlib

        geocodeglib_ver = str(GeocodeGlib._version)
    except Exception:
        geocodeglib_ver = _("not found")

    try:
        import bsddb3 as bsddb

        bsddb_str = bsddb.__version__
        bsddb_db_str = (
            str(bsddb.db.version()).replace(", ", ".").replace("(", "").replace(")", "")
        )
        bsddb_location_str = bsddb.__file__
    except:
        try:
            import berkeleydb as bsddb

            bsddb_str = bsddb.__version__
            bsddb_db_str = (
                str(bsddb.db.version())
                .replace(", ", ".")
                .replace("(", "")
                .replace(")", "")
            )
            bsddb_location_str = bsddb.__file__
        except:
            bsddb_str = "not found"
            bsddb_db_str = "not found"
            bsddb_location_str = "not found"

    try:
        import sqlite3

        sqlite3_py_version_str = sqlite3.version
        sqlite3_version_str = sqlite3.sqlite_version
        sqlite3_location_str = sqlite3.__file__
    except:
        sqlite3_version_str = "not found"
        sqlite3_py_version_str = "not found"
        sqlite3_location_str = "not found"

    try:
        from .gen.const import VERSION

        gramps_str = VERSION
    except:
        gramps_str = "not found"

    if hasattr(os, "uname"):
        kernel = os.uname()[2]
    else:
        kernel = None

    lang_str = get_env_var("LANG", "not set")
    language_str = get_env_var("LANGUAGE", "not set")
    grampsi18n_str = get_env_var("GRAMPSI18N", "not set")
    grampshome_str = get_env_var("GRAMPSHOME", "not set")
    grampsdir_str = get_env_var("GRAMPSDIR", "not set")
    gramps_resources_str = get_env_var("GRAMPS_RESOURCES", "not set")

    try:
        dotversion_str = Popen(["dot", "-V"], stderr=PIPE).communicate(input=None)[1]
        if isinstance(dotversion_str, bytes) and sys.stdin.encoding:
            dotversion_str = dotversion_str.decode(sys.stdin.encoding)
        if dotversion_str:
            dotversion_str = dotversion_str.replace("\n", "")[23:27]
    except:
        dotversion_str = "Graphviz not in system PATH"

    try:
        if win():
            try:
                gsversion_str = Popen(
                    ["gswin32c", "--version"], stdout=PIPE
                ).communicate(input=None)[0]
            except:
                gsversion_str = Popen(
                    ["gswin64c", "--version"], stdout=PIPE
                ).communicate(input=None)[0]
        else:
            gsversion_str = Popen(["gs", "--version"], stdout=PIPE).communicate(
                input=None
            )[0]
        if isinstance(gsversion_str, bytes) and sys.stdin.encoding:
            gsversion_str = gsversion_str.decode(sys.stdin.encoding)
        if gsversion_str:
            gsversion_str = gsversion_str.replace("\n", "")
    except:
        gsversion_str = "Ghostscript not in system PATH"

    os_path = get_env_var("PATH", "not set")
    os_path = os_path.split(os.pathsep)

    print("Gramps Settings:")
    print("----------------")
    print(" gramps    : %s" % gramps_str)
    print(" o.s.      : %s" % platform.system())
    if kernel:
        print(" kernel    : %s" % kernel)
    print("")
    print("Required:")
    print("---------")
    print(" Python    : %s" % py_str)
    print(" Gtk++     : %s" % gtkver_str)
    print(" pygobject : %s" % pygobjectver_str)
    print(" Cairo     : %s" % cairover_str)
    print(" pycairo   : %s" % pycairover_str)
    print(" Pango     : %s" % pangover_str)
    print(" PangoCairo: %s" % pangocairo_str)
    print("")
    print("Recommended:")
    print("------------")
    print(" osmgpsmap : %s" % osmgpsmap_str)
    print(" Graphviz  : %s" % dotversion_str)
    print(" Ghostscr. : %s" % gsversion_str)
    print(" ICU       : %s" % icu_str)
    print(" PyICU     : %s" % pyicu_str)
    print("")
    print("Optional:")
    print("---------")
    print(" Gspell     :", gspell_ver)
    print(" RCS        :", rcs_ver)
    print(" PILLOW     :", pil_ver)
    print(" GExiv2     : %s" % gexiv2_str)
    print(" Exiv2 lib. : %s" % vers_str)
    print(" geocodeglib: %s" % geocodeglib_ver)
    print("")
    print("Environment settings:")
    print("---------------------")
    print(" LANG      : %s" % lang_str)
    print(" LANGUAGE  : %s" % language_str)
    print(" GRAMPSI18N: %s" % grampsi18n_str)
    print(" GRAMPSHOME: %s" % grampshome_str)
    print(" GRAMPSDIR : %s" % grampsdir_str)
    if __debug__:
        print(" GRAMPS_RESOURCES : %s" % gramps_resources_str)
    print(" PYTHONPATH:")
    for folder in sys.path:
        print("   ", folder)
    print("")
    print("System PATH env variable:")
    print("-------------------------")
    for folder in os_path:
        print("    ", folder)
    print("")
    print("Databases:")
    print("-------------------------")
    print(" bsddb     :")
    print("     version     : %s" % bsddb_str)
    print("     db version  : %s" % bsddb_db_str)
    print("     location    : %s" % bsddb_location_str)
    print(" sqlite3   :")
    print("     version     : %s" % sqlite3_version_str)
    print("     py version  : %s" % sqlite3_py_version_str)
    print("     location    : %s" % sqlite3_location_str)
    print("")


def run():
    error = []

    try:
        build_user_paths()
    except OSError as msg:
        error += [(_("Configuration error:"), str(msg))]
        return error
    except msg:
        LOG.error("Could not read configuration.", exc_info=True)
        return [(_("Could not read configuration"), str(msg))]

    if not mime_type_is_defined(APP_GRAMPS):
        error += [
            (
                _("Configuration error:"),
                _(
                    "A definition for the media type %s could not "
                    "be found \n\n Possibly the installation of Gramps "
                    "was incomplete. Make sure something to handle the "
                    "media types of Gramps is installed."
                )
                % APP_GRAMPS,
            )
        ]

    # we start with parsing the arguments to determine if we have a cli or a
    # gui session

    if "-v" in sys.argv or "--version" in sys.argv:
        show_settings()
        return error

    from .cli.argparser import ArgParser

    argv_copy = sys.argv[:]
    argpars = ArgParser(argv_copy)

    # if in safe mode we should point the db dir back to the original dir.
    # It is ok to import config here, 'Defaults' command had its chance...
    if "SAFEMODE" in os.environ:
        config.set("database.path", os.path.join(ORIG_HOME_DIR, "grampsdb"))

    # On windows the fontconfig handler is a better choice
    if win() and ("PANGOCAIRO_BACKEND" not in os.environ):
        os.environ["PANGOCAIRO_BACKEND"] = "fontconfig"

    # Calls to LOG must be after setup_logging() and ArgParser()
    LOG = logging.getLogger(".locale")
    LOG.debug("Encoding: %s", glocale.encoding)
    LOG.debug("Translating Gramps to %s", glocale.language[0])
    LOG.debug("Collation Locale: %s", glocale.collation)
    LOG.debug("Date/Time Locale: %s", glocale.calendar)

    if "LANG" in os.environ:
        LOG.debug("Using LANG: %s" % get_env_var("LANG"))
    else:
        LOG.debug("environment: LANG is not defined")
    if "LANGUAGE" in os.environ:
        LOG.debug("Using LANGUAGE: %s" % get_env_var("LANGUAGE"))
    else:
        LOG.debug("environment: LANGUAGE is not defined")

    if argpars.need_gui():
        LOG.debug("A GUI is needed. Set one up.")
        try:
            from .gui.grampsgui import startgramps

            # no DISPLAY is a RuntimeError in an older pygtk (e.g. 2.17 in Fedora 14)
        except RuntimeError as msg:
            error += [(_("Configuration error:"), str(msg))]
            return error
        startgramps(error, argpars)
    else:
        # CLI use of Gramps
        argpars.print_help()
        argpars.print_usage()
        from .cli.grampscli import startcli

        startcli(error, argpars)


def main():
    if "GRAMPS_RESOURCES" not in os.environ:
        resource_path, filename = os.path.split(os.path.abspath(__file__))
        resource_path, dirname = os.path.split(resource_path)
        os.environ["GRAMPS_RESOURCES"] = resource_path
    errors = run()
    if errors and isinstance(errors, list):
        for error in errors:
            logging.warning(error[0] + error[1])

    sys.stdout.close()
    sys.stderr.close()


if __name__ == "__main__":
    main()



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2009       Nick Hall
# Copyright (C) 2010       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Person Tree View
"""

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.views.listview import TEXT, MARKUP, ICON
from gramps.plugins.lib.libpersonview import BasePersonView
from gramps.gui.views.treemodels.peoplemodel import PersonTreeModel
from gramps.gen.lib import Name, Person, Surname
from gramps.gen.errors import WindowActiveError
from gramps.gui.editors import EditPerson
from gramps.gen.utils.db import preset_name

# -------------------------------------------------------------------------
#
# Internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PersonTreeView
#
# -------------------------------------------------------------------------
class PersonTreeView(BasePersonView):
    """
    A hierarchical view of the people based on family name.
    """

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        BasePersonView.__init__(
            self,
            pdata,
            dbstate,
            uistate,
            _("People Tree View"),
            PersonTreeModel,
            nav_group=nav_group,
        )

    def get_viewtype_stock(self):
        """
        Override the default icon.  Set for hierarchical view.
        """
        return "gramps-tree-group"

    def define_actions(self):
        """
        Define actions for the popup menu specific to the tree view.
        """
        BasePersonView.define_actions(self)

        self.action_list.extend(
            [
                ("OpenAllNodes", self.open_all_nodes),
                ("CloseAllNodes", self.close_all_nodes),
            ]
        )

    additional_ui = BasePersonView.additional_ui[:]
    additional_ui.append(  # Defines the UI string for UIManager
        """
      <section id="PopUpTree">
        <item>
          <attribute name="action">win.OpenAllNodes</attribute>
          <attribute name="label" translatable="yes">"""
        """Expand all Nodes</attribute>
        </item>
        <item>
          <attribute name="action">win.CloseAllNodes</attribute>
          <attribute name="label" translatable="yes">"""
        """Collapse all Nodes</attribute>
        </item>
      </section>
    """
    )

    def add(self, *obj):
        person = Person()

        # attempt to get the current surname
        (model, pathlist) = self.selection.get_selected_rows()
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        basepers = None
        if len(pathlist) == 1:
            path = pathlist[0]
            pathids = path.get_indices()
            if len(pathids) == 1:
                path = Gtk.TreePath((pathids[0], 0))
            iter_ = model.get_iter(path)
            handle = model.get_handle_from_iter(iter_)
            basepers = self.dbstate.db.get_person_from_handle(handle)
        if basepers:
            preset_name(basepers, name)
        person.set_primary_name(name)
        try:
            EditPerson(self.dbstate, self.uistate, [], person)
        except WindowActiveError:
            pass

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  Stephane Charette
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
# "Objects with a certain reference count"
# -------------------------------------------------------------------------
class HasReferenceCountBase(Rule):
    """Objects with a reference count of <count>."""

    labels = [_("Reference count must be:"), _("Reference count:")]
    name = "Objects with a reference count of <count>"
    description = "Matches objects with a certain reference count"
    category = _("General filters")

    def prepare(self, db, user):
        # things we want to do just once, not for every handle
        if self.list[0] == "less than":
            self.count_type = 0
        elif self.list[0] == "greater than":
            self.count_type = 2
        else:
            self.count_type = 1  # "equal to"

        self.userSelectedCount = int(self.list[1])

    def apply(self, db, obj):
        handle = obj.get_handle()
        count = 0
        for item in db.find_backlink_handles(handle):
            count += 1

        if self.count_type == 0:  # "less than"
            return count < self.userSelectedCount
        elif self.count_type == 2:  # "greater than"
            return count > self.userSelectedCount
        # "equal to"
        return count == self.userSelectedCount



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python libraries
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.datehandler import get_date
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.utils.lds import TEMPLES


# -------------------------------------------------------------------------
#
# LdsModel
#
# -------------------------------------------------------------------------
class LdsModel(Gtk.ListStore):
    _HANDLE_COL = 6

    def __init__(self, lds_list, db):
        Gtk.ListStore.__init__(self, str, str, str, str, str, bool, bool, object)

        for lds_ord in lds_list:
            self.append(
                row=[
                    lds_ord.type2str(),
                    get_date(lds_ord),
                    lds_ord.status2str(),
                    TEMPLES.name(lds_ord.get_temple()),
                    place_displayer.display_event(db, lds_ord),
                    lds_ord.has_citations(),
                    lds_ord.get_privacy(),
                    lds_ord,
                ]
            )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2009  Brian G. Matherly
# Copyright (C) 2008       James Friedmann <jfriedmannj@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2015-2016  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A collection of utilities to aid in the generation of reports.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import os

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ...datehandler import get_date
from ...display.place import displayer as _pd
from ...utils.file import media_path_full
from ..docgen import IndexMark, INDEX_TYPE_ALP


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# -------------------------------------------------------------------------
#
#  Convert points to cm and back
#
# -------------------------------------------------------------------------
def pt2cm(pt_):
    """
    Convert points to centimeters. Fonts are typically specified in points,
    but the :class:`.BaseDoc` classes use centimeters.

    :param pt_: points
    :type pt_: float or int
    :returns: equivalent units in centimeters
    :rtype: float
    """
    return pt_ / 28.3465


def cm2pt(cm_):
    """
    Convert centimeters to points. Fonts are typically specified in points,
    but the :class:`.BaseDoc` classes use centimeters.

    :param cm_: centimeters
    :type cm_: float or int
    :returns: equivalent units in points
    :rtype: float
    """
    return cm_ * 28.3465


def rgb_color(color):
    """
    Convert color value from 0-255 integer range into 0-1 float range.

    :param color: list or tuple of integer values for red, green, and blue
    :type color: int
    :returns: (red, green, blue) tuple of floating point color values
    :rtype: 3-tuple
    """
    red = float(color[0]) / 255.0
    green = float(color[1]) / 255.0
    blue = float(color[2]) / 255.0
    return (red, green, blue)


# -------------------------------------------------------------------------
#
#  Roman numbers
#
# -------------------------------------------------------------------------
def roman(num):
    """Integer to Roman numeral converter for 0 < num < 4000"""
    if not isinstance(num, int):
        return "?"
    if not 0 < num < 4000:
        return "?"
    vals = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    nums = ("M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I")
    retval = ""
    for i, value in enumerate(vals):
        amount = int(num / value)
        retval += nums[i] * amount
        num -= value * amount
    return retval


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def place_name(dbase, place_handle):
    """returns a place string given a handle"""
    if place_handle:
        place = dbase.get_place_from_handle(place_handle)
        name = _pd.display(dbase, place)
    else:
        name = ""
    return str(name)


# -------------------------------------------------------------------------
#
# Functions commonly used in reports
#
# -------------------------------------------------------------------------
def insert_image(
    database,
    doc,
    photo,
    user,
    w_cm=4.0,
    h_cm=4.0,
    alt="",
    style_name=None,
    align="right",
):
    """
    Insert pictures of a person into the document.

    :type photo: :class:`~.mediaref.MediaRef`
    :type user:  :class:`.gen.user.User`
    :param alt: an alternative text to use. Useful for eg html reports
    :param style_name: style to use for the "alternative text" or captions
    :param align: image alignment: 'left', 'right', 'center', or 'single'
    """

    object_handle = photo.get_reference_handle()
    media = database.get_media_from_handle(object_handle)
    mime_type = media.get_mime_type()
    if mime_type and mime_type.startswith("image"):
        filename = media_path_full(database, media.get_path())
        if os.path.exists(filename):
            doc.add_media(
                filename,
                align,
                w_cm,
                h_cm,
                alt=alt,
                style_name=style_name,
                crop=photo.get_rectangle(),
            )
        else:
            no_file = _("File does not exist")
            user.warn(
                _("Could not add photo to page"),
                _("%(str1)s: %(str2)s") % {"str1": filename, "str2": no_file},
            )


# -------------------------------------------------------------------------
#
# find_spouse
#
# -------------------------------------------------------------------------
def find_spouse(person, family):
    """find the spouse of a person"""
    spouse_handle = None
    if family:
        if person.get_handle() == family.get_father_handle():
            spouse_handle = family.get_mother_handle()
        else:
            spouse_handle = family.get_father_handle()
    return spouse_handle


# -------------------------------------------------------------------------
#
# find_marriage
#
# -------------------------------------------------------------------------
def find_marriage(database, family):
    """find the marriage of a family"""
    for event_ref in family.get_event_ref_list():
        event = database.get_event_from_handle(event_ref.ref)
        if event and event.type.is_marriage() and event_ref.role.is_family():
            return event
    return None


# -------------------------------------------------------------------------
#
# Indexing function
#
# -------------------------------------------------------------------------
def get_person_mark(dbase, person):
    """
    Return a IndexMark that can be used to index a person in a report

    :param dbase: the Gramps database instance
    :param person: the key is for
    """
    if not person:
        return None

    name = person.get_primary_name().get_name()
    birth = " "
    death = " "
    key = ""

    birth_ref = person.get_birth_ref()
    if birth_ref:
        birth_event = dbase.get_event_from_handle(birth_ref.ref)
        birth = get_date(birth_event)

    death_ref = person.get_death_ref()
    if death_ref:
        death_event = dbase.get_event_from_handle(death_ref.ref)
        death = get_date(death_event)

    if birth == death == " ":
        key = name
    else:
        key = "%s (%s - %s)" % (name, birth, death)

    return IndexMark(key, INDEX_TYPE_ALP)


# -------------------------------------------------------------------------
#
# Address String
#
# -------------------------------------------------------------------------
def get_address_str(addr):
    """
    Return a string that combines the elements of an address

    :param addr: the Gramps address instance
    """
    addr_str = ""
    elems = [
        addr.get_street(),
        addr.get_locality(),
        addr.get_city(),
        addr.get_county(),
        addr.get_state(),
        addr.get_country(),
        addr.get_postal_code(),
        addr.get_phone(),
    ]

    for info in elems:
        if info:
            if addr_str == "":
                addr_str = info
            else:
                # Translators: needed for Arabic, ignore otherwise
                addr_str = _("%(str1)s, %(str2)s") % {"str1": addr_str, "str2": info}
    return addr_str


# -------------------------------------------------------------------------
#
# People Filters
#
# -------------------------------------------------------------------------
def get_person_filters(person, include_single=True, name_format=None):
    """
    Return a list of filters that are relevant for the given person

    :param person: the person the filters should apply to.
    :type person: :class:`~.person.Person`
    :param include_single: include a filter to include the single person
    :type include_single: boolean
    :param name_format: optional format to control display of person's name
    :type name_format: None or int
    """
    from ...filters import GenericFilter, rules, CustomFilters, DeferredFilter
    from ...display.name import displayer as name_displayer

    if person:
        real_format = name_displayer.get_default_format()
        if name_format is not None:
            name_displayer.set_default_format(name_format)
        name = name_displayer.display(person)
        name_displayer.set_default_format(real_format)
        gramps_id = person.get_gramps_id()
    else:
        # Do this in case of command line options query (show=filter)
        name = _("PERSON")
        gramps_id = ""

    if include_single:
        filt_id = GenericFilter()
        filt_id.set_name(name)
        filt_id.add_rule(rules.person.HasIdOf([gramps_id]))

    all_people = DeferredFilter(_T_("Entire Database"), None)
    all_people.add_rule(rules.person.Everyone([]))

    # feature request 2356: avoid genitive form
    des = DeferredFilter(_T_("Descendants of %s"), name)
    des.add_rule(rules.person.IsDescendantOf([gramps_id, 1]))

    # feature request 2356: avoid genitive form
    d_fams = DeferredFilter(_T_("Descendant Families of %s"), name)
    d_fams.add_rule(rules.person.IsDescendantFamilyOf([gramps_id, 1]))

    # feature request 2356: avoid genitive form
    ans = DeferredFilter(_T_("Ancestors of %s"), name)
    ans.add_rule(rules.person.IsAncestorOf([gramps_id, 1]))

    com = DeferredFilter(_T_("People with common ancestor with %s"), name)
    com.add_rule(rules.person.HasCommonAncestorWith([gramps_id]))

    if include_single:
        the_filters = [filt_id, all_people, des, d_fams, ans, com]
    else:
        the_filters = [all_people, des, d_fams, ans, com]
    the_filters.extend(CustomFilters.get_filters("Person"))
    return the_filters


# -------------------------------------------------------------------------
#
# Family Filters
#
# -------------------------------------------------------------------------
def get_family_filters(database, family, include_single=True, name_format=None):
    """
    Return a list of filters that are relevant for the given family

    :param database: The database that the family is in.
    :type database: DbBase
    :param family: the family the filters should apply to.
    :type family: :class:`~.family.Family`
    :param include_single: include a filter to include the single family
    :type include_single: boolean
    :param name_format: optional format to control display of person's name
    :type name_format: None or int
    """
    from ...filters import (
        GenericFilterFactory,
        rules,
        CustomFilters,
        DeferredFamilyFilter,
    )
    from ...display.name import displayer as name_displayer

    if family:
        real_format = name_displayer.get_default_format()
        if name_format is not None:
            name_displayer.set_default_format(name_format)
        fhandle = family.get_father_handle()
        if fhandle:
            father = database.get_person_from_handle(fhandle)
            father_name = name_displayer.display(father)
        else:
            father_name = _("unknown father")
        mhandle = family.get_mother_handle()
        if mhandle:
            mother = database.get_person_from_handle(mhandle)
            mother_name = name_displayer.display(mother)
        else:
            mother_name = _("unknown mother")
        gramps_id = family.get_gramps_id()
        name = _("%(father_name)s and %(mother_name)s (%(family_id)s)") % {
            "father_name": father_name,
            "mother_name": mother_name,
            "family_id": gramps_id,
        }
        name_displayer.set_default_format(real_format)
    else:
        # Do this in case of command line options query (show=filter)
        name = _("FAMILY")
        gramps_id = ""

    if include_single:
        FilterClass = GenericFilterFactory("Family")
        filt_id = FilterClass()
        filt_id.set_name(name)
        filt_id.add_rule(rules.family.HasIdOf([gramps_id]))

    all_families = DeferredFamilyFilter(_T_("Every family"), None)
    all_families.add_rule(rules.family.AllFamilies([]))

    # feature request 2356: avoid genitive form
    d_fams = DeferredFamilyFilter(_T_("Descendant Families of %s"), name)
    d_fams.add_rule(rules.family.IsDescendantOf([gramps_id, 1]))

    # feature request 2356: avoid genitive form
    ans = DeferredFamilyFilter(_T_("Ancestor Families of %s"), name)
    ans.add_rule(rules.family.IsAncestorOf([gramps_id, 1]))

    if include_single:
        the_filters = [filt_id, all_families, d_fams, ans]
    else:
        the_filters = [all_families, d_fams, ans]
    the_filters.extend(CustomFilters.get_filters("Family"))
    return the_filters



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Family Tree Processing/MergeCitations"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# GNOME libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.utils.string import conf_strings
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gui.utils import ProgressMeter
from gramps.gui.plug import tool
from gramps.gui.dialog import OkDialog
from gramps.gui.display import display_help
from gramps.gen.datehandler import get_date
from gramps.gui.managedwindow import ManagedWindow
from gramps.gen.merge import MergeCitationQuery

from gramps.gui.glade import Glade
from gramps.gen.db import DbTxn
from gramps.gen.lib import Person, Family, Event, Place, Media, Citation, Repository
from gramps.gen.errors import MergeError

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
ALL_FIELDS = 0
IGNORE_DATE = 1
IGNORE_CONFIDENCE = 2
IGNORE_BOTH = 3

_val2label = {
    ALL_FIELDS: _("Match on Page/Volume, Date and Confidence"),
    IGNORE_DATE: _("Ignore Date"),
    IGNORE_CONFIDENCE: _("Ignore Confidence"),
    IGNORE_BOTH: _("Ignore Date and Confidence"),
}

WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Merge_citations", "manual")


# -------------------------------------------------------------------------
#
# The Actual tool.
#
# -------------------------------------------------------------------------
class MergeCitations(tool.BatchTool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.user = user

        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate
        self.set_window(Gtk.Window(), Gtk.Label(), "")

        tool.BatchTool.__init__(self, dbstate, user, options_class, name)

        if not self.fail:
            uistate.set_busy_cursor(True)
            self.run()
            uistate.set_busy_cursor(False)

    def run(self):
        top = Glade(toplevel="mergecitations", also_load=["liststore1"])

        # retrieve options
        fields = self.options.handler.options_dict["fields"]
        dont_merge_notes = self.options.handler.options_dict["dont_merge_notes"]

        my_menu = Gtk.ListStore(str, object)
        for val in sorted(_val2label):
            my_menu.append([_val2label[val], val])

        self.notes_obj = top.get_object("notes")
        self.notes_obj.set_active(dont_merge_notes)
        self.notes_obj.show()

        self.menu = top.get_object("menu")
        self.menu.set_model(my_menu)
        self.menu.set_active(fields)

        window = top.toplevel
        window.set_transient_for(self.user.uistate.window)
        window.show()
        #        self.set_window(window, top.get_object('title'),
        #                        _('Merge citations'))
        self.set_window(
            window,
            top.get_object("title2"),
            _(
                "Notes, media objects and data-items of matching "
                "citations will be combined."
            ),
        )
        self.setup_configs("interface.mergecitations", 700, 230)

        top.connect_signals(
            {
                "on_merge_ok_clicked": self.on_merge_ok_clicked,
                "destroy_passed_object": self.cancel,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_merge_event": self.close,
                "on_delete_event": self.close,
            }
        )

        self.show()

    def cancel(self, obj):
        """
        on cancel, update the saved values of the options.
        """
        fields = self.menu.get_model()[self.menu.get_active()][1]
        dont_merge_notes = int(self.notes_obj.get_active())
        LOG.debug("cancel fields %d dont_merge_notes %d" % (fields, dont_merge_notes))

        self.options.handler.options_dict["fields"] = fields
        self.options.handler.options_dict["dont_merge_notes"] = dont_merge_notes
        # Save options
        self.options.handler.save_options()

        self.close(obj)

    def build_menu_names(self, obj):
        return (_("Tool settings"), _("Merge citations tool"))

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""

        display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)

    def on_merge_ok_clicked(self, obj):
        """
        Performs the actual merge of citations
        (Derived from ExtractCity)
        """
        fields = self.menu.get_model()[self.menu.get_active()][1]
        dont_merge_notes = int(self.notes_obj.get_active())
        LOG.debug("fields %d dont_merge_notes %d" % (fields, dont_merge_notes))

        self.options.handler.options_dict["fields"] = fields
        self.options.handler.options_dict["dont_merge_notes"] = dont_merge_notes
        # Save options
        self.options.handler.save_options()

        self.progress = ProgressMeter(_("Checking Sources"), "", parent=self.window)
        self.progress.set_pass(
            _("Looking for citation fields"), self.db.get_number_of_citations()
        )

        db = self.dbstate.db

        db.disable_signals()
        num_merges = 0
        for handle in db.iter_source_handles():
            dict = {}
            citation_handle_list = list(db.find_backlink_handles(handle))
            for class_name, citation_handle in citation_handle_list:
                if class_name != Citation.__name__:
                    raise MergeError(
                        "Encountered an object of type %s "
                        "that has a citation reference." % class_name
                    )

                citation = db.get_citation_from_handle(citation_handle)
                if citation is None:
                    continue
                key = citation.get_page()
                if fields != IGNORE_DATE and fields != IGNORE_BOTH:
                    key += "\n" + get_date(citation)
                if fields != IGNORE_CONFIDENCE and fields != IGNORE_BOTH:
                    key += "\n" + conf_strings[citation.get_confidence_level()]
                if key in dict and (
                    not dont_merge_notes or len(citation.note_list) == 0
                ):
                    citation_match_handle = dict[key]
                    citation_match = db.get_citation_from_handle(citation_match_handle)
                    try:
                        query = MergeCitationQuery(
                            self.dbstate, citation_match, citation
                        )
                        query.execute()
                    except AssertionError:
                        print(
                            "Tool/Family Tree processing/MergeCitations",
                            "citation1 gramps_id",
                            citation_match.get_gramps_id(),
                            "citation2 gramps_id",
                            citation.get_gramps_id(),
                            "citation backlink handles",
                            list(db.find_backlink_handles(citation.get_handle())),
                        )
                    num_merges += 1
                elif not dont_merge_notes or len(citation.note_list) == 0:
                    dict[key] = citation_handle
                self.progress.step()
        db.enable_signals()
        db.request_rebuild()
        self.progress.close()
        OkDialog(
            _("Number of merges done"),
            # Translators: leave all/any {...} untranslated
            ngettext(
                "{number_of} citation merged",
                "{number_of} citations merged",
                num_merges,
            ).format(number_of=num_merges),
            parent=self.window,
        )
        self.close(obj)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class MergeCitationsOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "fields": 1,
            "dont_merge_notes": 0,
        }
        self.options_help = {
            "dont_merge_notes": (
                "=0/1",
                "Whether to merge citations if they have notes",
                ["Merge citations with notes", "Do not merge citations with notes"],
                False,
            ),
            "fields": ("=num", "Threshold for matching", "Integer number"),
        }



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib.nameorigintype import NameOriginType


# -------------------------------------------------------------------------
#
# HasNameOf
#
# -------------------------------------------------------------------------
class HasNameOf(Rule):
    """Rule that checks for full or partial name matches"""

    labels = [
        _("Given name:"),
        _("Full Family name:"),
        _("Title:", "person"),
        _("Suffix:"),
        _("Call Name:"),
        _("Nick Name:"),
        _("Prefix:"),
        _("Single Surname:"),
        _("Connector"),
        _("Patronymic:"),
        _("Family Nick Name:"),
    ]
    name = _("People with the <name>")
    description = _("Matches people with a specified (partial) name")
    category = _("General filters")
    allow_regex = True

    def apply(self, db, person):
        for name in [person.get_primary_name()] + person.get_alternate_names():
            if self.match_name(name):
                return True
        return False

    def match_name(self, name):
        if self.list[0] and not self.match_substring(0, name.get_first_name()):
            return False
        elif self.list[1] and not self.match_substring(1, name.get_surname()):
            return False
        elif self.list[2] and not self.match_substring(2, name.get_title()):
            return False
        elif self.list[3] and not self.match_substring(3, name.get_suffix()):
            return False
        elif self.list[4] and not self.match_substring(4, name.get_call_name()):
            return False
        elif self.list[5] and not self.match_substring(5, name.get_nick_name()):
            return False
        elif self.list[10] and not self.match_substring(
            10, name.get_family_nick_name()
        ):
            return False
        else:
            for surn in name.get_surname_list():
                if self.match_surname(surn):
                    return True
        return False

    def match_surname(self, surn):
        if self.list[6] and not self.match_substring(6, surn.get_prefix()):
            return False
        if self.list[7] and not self.match_substring(7, surn.get_surname()):
            return False
        if self.list[8] and not self.match_substring(8, surn.get_connector()):
            return False
        if surn.get_origintype().value == NameOriginType.PATRONYMIC:
            if self.list[9] and not self.match_substring(9, surn.get_surname()):
                return False
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide merge capabilities for events.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import Person, Family, Note
from ..db import DbTxn
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..errors import MergeError


# -------------------------------------------------------------------------
#
# MergeEventQuery
#
# -------------------------------------------------------------------------
class MergeEventQuery:
    """
    Create database query to merge two events.
    """

    def __init__(self, dbstate, phoenix, titanic):
        self.database = dbstate.db
        self.phoenix = phoenix
        self.titanic = titanic

    def execute(self):
        """
        Merges two events into a single event.
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        self.phoenix.merge(self.titanic)

        with DbTxn(_("Merge Event Objects"), self.database) as trans:
            self.database.commit_event(self.phoenix, trans)
            for class_name, handle in self.database.find_backlink_handles(old_handle):
                if class_name == Person.__name__:
                    person = self.database.get_person_from_handle(handle)
                    assert person.has_handle_reference("Event", old_handle)
                    bri = person.birth_ref_index
                    dri = person.death_ref_index
                    person.replace_handle_reference("Event", old_handle, new_handle)
                    if person.birth_ref_index != bri and person.birth_ref_index == -1:
                        for index, ref in enumerate(person.get_event_ref_list()):
                            event = self.database.get_event_from_handle(ref.ref)
                            if event.type.is_birth() and ref.role.is_primary():
                                person.birth_ref_index = index
                                break
                    if person.death_ref_index != dri and person.death_ref_index == -1:
                        for index, ref in enumerate(person.get_event_ref_list()):
                            event = self.database.get_event_from_handle(ref.ref)
                            if event.type.is_death() and ref.role.is_primary():
                                person.death_ref_index = index
                                break
                    self.database.commit_person(person, trans)
                elif class_name == Family.__name__:
                    family = self.database.get_family_from_handle(handle)
                    assert family.has_handle_reference("Event", old_handle)
                    family.replace_handle_reference("Event", old_handle, new_handle)
                    self.database.commit_family(family, trans)
                elif class_name == Note.__name__:
                    note = self.database.get_note_from_handle(handle)
                    assert note.has_handle_reference("Event", old_handle)
                    note.replace_handle_reference("Event", old_handle, new_handle)
                    self.database.commit_note(note, trans)
                else:
                    raise MergeError(
                        "Encounter an object of type %s that has "
                        "an event reference." % class_name
                    )
            self.database.remove_event(old_handle, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Benny Malengier
#
# based on undoablebuffer Copyright (C) 2009  Florian Heinle
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["UndoableEntry"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------

import warnings
import logging

_LOG = logging.getLogger(".widgets.undoableentry")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .undoablebuffer import Stack
from gramps.gen.const import GRAMPS_LOCALE as glocale

# table for skipping illegal control chars
INVISIBLE = dict.fromkeys(list(range(32)) + [0x202D, 0x202E])


class UndoableInsertEntry:
    """something that has been inserted into our Gtk.editable"""

    def __init__(self, text, length, position):
        self.offset = position
        self.text = text
        # unicode char can have length > 1 as it points in the buffer
        charlength = len(text)
        self.length = charlength
        if charlength > 1 or self.text in ("\r", "\n", " "):
            self.mergeable = False
        else:
            self.mergeable = True


class UndoableDeleteEntry:
    """something that has been deleted from our textbuffer"""

    def __init__(self, editable, start, end):
        self.text = editable.get_chars(start, end)
        self.start = start
        self.end = end
        # need to find out if backspace or delete key has been used
        # so we don't mess up during redo
        insert = editable.get_position()
        if insert <= start:
            self.delete_key_used = True
        else:
            self.delete_key_used = False
        if self.end - self.start > 1 or self.text in ("\r", "\n", " "):
            self.mergeable = False
        else:
            self.mergeable = True


class UndoableEntry(Gtk.Entry, Gtk.Editable):
    """
    The UndoableEntry is an Entry subclass with additional features.

    Additional features:
      - Undo and Redo on CTRL-Z/CTRL-SHIFT-Z

      - ltr_mode (forces the field to always be left to right, useful for GPS
        coordinates and similar numbers that might contain RTL characters.
        See set_ltr_mode.
    """

    __gtype_name__ = "UndoableEntry"

    insertclass = UndoableInsertEntry
    deleteclass = UndoableDeleteEntry

    # how many undo's are remembered
    undo_stack_size = 50

    def __init__(self):
        self.undo_stack = Stack(self.undo_stack_size)
        self.redo_stack = []
        self.not_undoable_action = False
        self.undo_in_progress = False
        self.ltr_mode = False
        Gtk.Entry.__init__(self)
        self.connect("delete-text", self._on_delete_text)
        self.connect("key-press-event", self._on_key_press_event)

    def set_text(self, text):
        with warnings.catch_warnings():
            # Suppress warnings.  See bug #8029.
            # https://bugzilla.gnome.org/show_bug.cgi?id=644927
            warnings.simplefilter("ignore")
            Gtk.Entry.set_text(self, text)
        self.reset()

    def _on_key_press_event(self, widget, event):
        """Signal handler.
        Handle formatting undo/redo key press.

        """
        keymap = Gdk.Keymap.get_for_display(Gdk.Display.get_default())
        primary = keymap.get_modifier_mask(Gdk.ModifierIntent.PRIMARY_ACCELERATOR)
        if (
            (Gdk.keyval_name(event.keyval) == "Z")
            and (event.get_state() & primary)
            and (event.get_state() & Gdk.ModifierType.SHIFT_MASK)
        ):
            self.redo()
            return True
        elif (Gdk.keyval_name(event.keyval) == "z") and (event.get_state() & primary):
            self.undo()
            return True

        return False

    def __empty_redo_stack(self):
        self.redo_stack = []

    def do_insert_text(self, text, length, position):
        def can_be_merged(prev, cur):
            """
            see if we can merge multiple inserts here

            will try to merge words or whitespace
            can't merge if prev and cur are not mergeable in the first place
            can't merge when user set the input bar somewhere else
            can't merge across word boundaries
            """

            WHITESPACE = (" ", "\t")
            if not cur.mergeable or not prev.mergeable:
                return False
            # offset is char offset, not byte, so length is the char length!
            elif cur.offset != (prev.offset + prev.length):
                return False
            elif cur.text in WHITESPACE and not prev.text in WHITESPACE:
                return False
            elif prev.text in WHITESPACE and not cur.text in WHITESPACE:
                return False
            return True

        text = text.translate(INVISIBLE)
        if self.ltr_mode:
            if position == 0:
                position = 1
            elif position >= self.get_text_length():
                position -= 1

        if not self.undo_in_progress:
            self.__empty_redo_stack()
        while not self.not_undoable_action:
            undo_action = self.insertclass(text, length, position)
            try:
                prev_insert = self.undo_stack.pop()
            except IndexError:
                self.undo_stack.append(undo_action)
                break
            if not isinstance(prev_insert, self.insertclass):
                self.undo_stack.append(prev_insert)
                self.undo_stack.append(undo_action)
                break
            if can_be_merged(prev_insert, undo_action):
                prev_insert.length += undo_action.length
                prev_insert.text += undo_action.text
                self.undo_stack.append(prev_insert)
            else:
                self.undo_stack.append(prev_insert)
                self.undo_stack.append(undo_action)
            break
        self.get_buffer().insert_text(position, text, len(text))
        return position + len(text)

    def _on_delete_text(self, editable, start, end):
        def can_be_merged(prev, cur):
            """
            see if we can merge multiple deletions here

            will try to merge words or whitespace
            can't merge if prev and cur are not mergeable in the first place
            can't merge if delete and backspace key were both used
            can't merge across word boundaries
            """

            WHITESPACE = (" ", "\t")
            if not cur.mergeable or not prev.mergeable:
                return False
            elif prev.delete_key_used != cur.delete_key_used:
                return False
            elif prev.start != cur.start and prev.start != cur.end:
                return False
            elif cur.text not in WHITESPACE and prev.text in WHITESPACE:
                return False
            elif cur.text in WHITESPACE and prev.text not in WHITESPACE:
                return False
            return True

        if self.ltr_mode:  # limit deletes to area between LRO/PDF
            if start == 0:
                start = 1
            elif start > self.get_text_length() - 1:
                start -= 1
            if end == 0:
                end = 1
            elif end > self.get_text_length() - 1:
                end -= 1
            elif end < 0:
                end = self.get_text_length() - 1

        while True:
            if not self.undo_in_progress:
                self.__empty_redo_stack()
            if self.not_undoable_action:
                break
            undo_action = self.deleteclass(self, start, end)
            try:
                prev_delete = self.undo_stack.pop()
            except IndexError:
                self.undo_stack.append(undo_action)
                break
            if not isinstance(prev_delete, self.deleteclass):
                self.undo_stack.append(prev_delete)
                self.undo_stack.append(undo_action)
                break
            if can_be_merged(prev_delete, undo_action):
                if prev_delete.start == undo_action.start:  # delete key used
                    prev_delete.text += undo_action.text
                    prev_delete.end += undo_action.end - undo_action.start
                else:  # Backspace used
                    prev_delete.text = "%s%s" % (undo_action.text, prev_delete.text)
                    prev_delete.start = undo_action.start
                self.undo_stack.append(prev_delete)
            else:
                self.undo_stack.append(prev_delete)
                self.undo_stack.append(undo_action)
            break
        self.get_buffer().delete_text(start, end - start)
        self.stop_emission_by_name("delete-text")
        return True

    def begin_not_undoable_action(self):
        """don't record the next actions

        toggles self.not_undoable_action"""
        self.not_undoable_action = True

    def end_not_undoable_action(self):
        """record next actions

        toggles self.not_undoable_action"""
        self.not_undoable_action = False

    def reset(self):
        """
        Resets buffer to initial state.
        """
        self.undo_stack = Stack(self.undo_stack_size)
        self.redo_stack[:] = []
        self.not_undoable_action = False
        self.undo_in_progress = False

    def undo(self):
        """undo inserts or deletions

        undone actions are being moved to redo stack"""
        if not self.undo_stack:
            return
        self.begin_not_undoable_action()
        self.undo_in_progress = True
        undo_action = self.undo_stack.pop()
        self.redo_stack.append(undo_action)
        if isinstance(undo_action, self.insertclass):
            self._undo_insert(undo_action)
        elif isinstance(undo_action, self.deleteclass):
            self._undo_delete(undo_action)
        else:
            self._handle_undo(undo_action)
        self.end_not_undoable_action()
        self.undo_in_progress = False

    def _undo_insert(self, undo_action):
        start = undo_action.offset
        stop = undo_action.offset + undo_action.length
        self.delete_text(start, stop)
        self.set_position(undo_action.offset)

    def _undo_delete(self, undo_action):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.insert_text(undo_action.text, undo_action.start)
        if undo_action.delete_key_used:
            self.set_position(undo_action.start)
        else:
            self.set_position(undo_action.end)

    def _handle_undo(self, undo_action):
        raise NotImplementedError

    def redo(self):
        """redo inserts or deletions

        redone actions are moved to undo stack"""
        if not self.redo_stack:
            return
        self.begin_not_undoable_action()
        self.undo_in_progress = True
        redo_action = self.redo_stack.pop()
        self.undo_stack.append(redo_action)
        if isinstance(redo_action, self.insertclass):
            self._redo_insert(redo_action)
        elif isinstance(redo_action, self.deleteclass):
            self._redo_delete(redo_action)
        else:
            self._handle_redo(redo_action)
        self.end_not_undoable_action()
        self.undo_in_progress = False

    def _redo_insert(self, redo_action):
        self.insert_text(redo_action.text, redo_action.offset)
        new_cursor_pos = redo_action.offset + redo_action.length
        self.set_position(new_cursor_pos)

    def _redo_delete(self, redo_action):
        start = redo_action.start
        stop = redo_action.end
        self.delete_text(start, stop)
        self.set_position(redo_action.start)

    def _handle_redo(self, redo_action):
        raise NotImplementedError

    def set_ltr_mode(self):
        """sets up the Entry to always be in LTR left to right even if some
        characters are RTL.
        This works by inserting the LRO/PDF Unicode Explicit Directional
        Override characters around the entry text.  These characters are then
        protected agains insert/delete operations.

        This call must be made before other text is inserted to the Entry.

        Note: we only enable this during rtl_local languages because it has a
        minor consequence; if cutting a field from this Entry with this mode
        enabled, the LRO/PDF characters may end up in the clipboard.  If pasted
        back into another UndoableEntry, this is ignored, but if pasted in
        another app it may be noticable.
        """
        if glocale.rtl_locale:
            self.get_buffer().set_text("\u202d\u202e", -1)
            self.ltr_mode = True

    def do_set_position(self, position):
        """In ltr_mode, this ensures that the cursor cannot be put outside
        the LRO/PDF characters on the ends of the buffer."""
        if position < 0:
            position = self.get_text_length()
        if self.ltr_mode:
            if position == 0:
                position = 1
            elif position == self.get_text_length():
                position -= 1
        Gtk.Editable.select_region(self, position, position)

    def get_text(self):
        """Used to remove the LRO/PDF characters when in ltr_mode."""
        text = Gtk.Entry.get_text(self)
        return text[1:-1] if self.ltr_mode else text



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Polish version 2007 by Piotr Czubaszek
# Updated in 2010 by Łukasz Rymarczyk

"""
Polish-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Polish parser
#
# -------------------------------------------------------------------------
class DateParserPL(DateParser):
    month_to_int = DateParser.month_to_int
    month_to_int["styczeń"] = 1
    month_to_int["sty"] = 1
    month_to_int["I"] = 1
    month_to_int["luty"] = 2
    month_to_int["lut"] = 2
    month_to_int["II"] = 2
    month_to_int["marzec"] = 3
    month_to_int["mar"] = 3
    month_to_int["III"] = 3
    month_to_int["kwiecień"] = 4
    month_to_int["kwi"] = 4
    month_to_int["IV"] = 4
    month_to_int["maj"] = 5
    month_to_int["V"] = 5
    month_to_int["czerwiec"] = 6
    month_to_int["cze"] = 6
    month_to_int["VI"] = 6
    month_to_int["lipiec"] = 7
    month_to_int["lip"] = 7
    month_to_int["VII"] = 7
    month_to_int["sierpień"] = 8
    month_to_int["sie"] = 8
    month_to_int["VIII"] = 8
    month_to_int["wrzesień"] = 9
    month_to_int["wrz"] = 9
    month_to_int["IX"] = 9
    month_to_int["październik"] = 10
    month_to_int["paź"] = 10
    month_to_int["X"] = 10
    month_to_int["listopad"] = 11
    month_to_int["lis"] = 11
    month_to_int["XI"] = 11
    month_to_int["grudzień"] = 12
    month_to_int["gru"] = 12
    month_to_int["XII"] = 12
    # Alternative forms: declined nouns
    month_to_int["stycznia"] = 1
    month_to_int["lutego"] = 2
    month_to_int["marca"] = 3
    month_to_int["kwietnia"] = 4
    month_to_int["maja"] = 5
    month_to_int["czerwca"] = 6
    month_to_int["lipca"] = 7
    month_to_int["sierpnia"] = 8
    month_to_int["września"] = 9
    month_to_int["października"] = 10
    month_to_int["listopada"] = 11
    month_to_int["grudnia"] = 12
    # Alternative forms: nouns without polish accent letters
    # (misspellings sometimes used in emails)
    month_to_int["styczen"] = 1
    month_to_int["kwiecien"] = 4
    month_to_int["sierpien"] = 8
    month_to_int["wrzesien"] = 9
    month_to_int["pazdziernik"] = 10
    month_to_int["grudzien"] = 12
    month_to_int["wrzesnia"] = 9
    month_to_int["pazdziernika"] = 10
    month_to_int["paz"] = 10

    modifier_to_int = {
        "przed": Date.MOD_BEFORE,
        "po": Date.MOD_AFTER,
        "około": Date.MOD_ABOUT,
        "ok.": Date.MOD_ABOUT,
        "circa": Date.MOD_ABOUT,
        "ca.": Date.MOD_ABOUT,
        # Alternative forms: misspellings sometimes used in emails
        "okolo": Date.MOD_ABOUT,
        "ok": Date.MOD_ABOUT,
        "od": Date.MOD_FROM,
        "do": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregoriański": Date.CAL_GREGORIAN,
        "greg.": Date.CAL_GREGORIAN,
        "juliański": Date.CAL_JULIAN,
        "jul.": Date.CAL_JULIAN,
        "hebrajski": Date.CAL_HEBREW,
        "hebr.": Date.CAL_HEBREW,
        "islamski": Date.CAL_ISLAMIC,
        "isl.": Date.CAL_ISLAMIC,
        "francuski republikański": Date.CAL_FRENCH,
        "franc.": Date.CAL_FRENCH,
        "perski": Date.CAL_PERSIAN,
        "szwedzki": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
        # Alternative forms: nouns without polish accent letters
        # (misspellings sometimes used in emails)
        "gregorianski": Date.CAL_GREGORIAN,
        "julianski": Date.CAL_JULIAN,
        "francuski republikanski": Date.CAL_FRENCH,
    }

    quality_to_int = {
        "szacowany": Date.QUAL_ESTIMATED,
        "szac.": Date.QUAL_ESTIMATED,
        "obliczony": Date.QUAL_CALCULATED,
        "obl.": Date.QUAL_CALCULATED,
    }

    bce = ["przed naszą erą", "przed Chrystusem", "p.n.e."] + DateParser.bce

    def init_strings(self):
        DateParser.init_strings(self)
        self._span = re.compile(
            r"(od)\s+(?P<start>.+)\s+(do)\s+(?P<stop>.+)", re.IGNORECASE
        )
        # Also handle a common mistakes
        self._range = re.compile(
            r"((?:po)?mi(?:ę|e)dzy)\s+(?P<start>.+)\s+(a)\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._text2 = re.compile(
            r"(\d+)?.?\s+?%s\s*((\d+)(/\d+)?)?" % self._mon_str, re.IGNORECASE
        )
        self._jtext2 = re.compile(
            r"(\d+)?.?\s+?%s\s*((\d+)(/\d+)?)?" % self._jmon_str, re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Polish display
#
# -------------------------------------------------------------------------
class DateDisplayPL(DateDisplay):
    """
    Polish language date display class.
    """

    long_months = (
        "",
        "Styczeń",
        "Luty",
        "Marzec",
        "Kwiecień",
        "Maj",
        "Czerwiec",
        "Lipiec",
        "Sierpień",
        "Wrzesień",
        "Październik",
        "Listopad",
        "Grudzień",
    )

    short_months = (
        "",
        "Sty",
        "Lut",
        "Mar",
        "Kwi",
        "Maj",
        "Cze",
        "Lip",
        "Sie",
        "Wrz",
        "Paź",
        "Lis",
        "Gru",
    )

    calendar = (
        "",
        "juliański",
        "hebrajski",
        "francuski republikański",
        "perski",
        "islamski",
        "swedish",
    )

    _mod_str = ("", "przed ", "po ", "ok. ", "", "", "")

    _qual_str = ("", "szacowany ", "obliczony ")

    _bce_str = "%s p.n.e."

    formats = (
        "RRRR-MM-DD (ISO)",
        "Numeryczny",
        "Miesiąc Dzień, Rok",
        "Miesiąc.Dzień.Rok",
        "Dzień Miesiąc Rok",
        "Dzień MieRzym Rok",
    )
    # this definition must agree with its "_display_gregorian" method

    roman_months = (
        "",
        "I",
        "II",
        "III",
        "IV",
        "V",
        "VI",
        "VII",
        "VIII",
        "IX",
        "X",
        "XI",
        "XII",
    )

    def _display_gregorian(self, date_val, **kwargs):
        """
        display gregorian calendar date in different format
        """
        # this must agree with its locale-specific "formats" definition
        year = self._slash_year(date_val[2], date_val[3])
        if self.format == 0:
            return self.display_iso(date_val)
        elif self.format == 1:
            # day.month_number.year
            if date_val[3]:
                return self.display_iso(date_val)
            else:
                if date_val[0] == date_val[1] == 0:
                    value = str(date_val[2])
                else:
                    value = self.dhformat.replace("%d", str(date_val[0]))
                    value = value.replace("%m", str(date_val[1]))
                    value = value.replace("%Y", str(date_val[2]))
        elif self.format == 2:
            # month_name day, year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s %s" % (self.long_months[date_val[1]], year)
            else:
                value = "%s %d, %s" % (self.long_months[date_val[1]], date_val[0], year)
        elif self.format == 3:
            # month_number. day. year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%d.%s" % (date_val[1], year)
            else:
                value = "%d.%d.%s" % (date_val[1], date_val[0], year)
        elif self.format == 4:
            # day month_name year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s %s" % (self.long_months[date_val[1]], year)
            else:
                value = "%d %s %s" % (date_val[0], self.long_months[date_val[1]], year)
        else:
            # day Roman_number_month year
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = year
                else:
                    value = "%s %s" % (self.roman_months[date_val[1]], year)
            else:
                value = "%d %s %s" % (date_val[0], self.roman_months[date_val[1]], year)
        if date_val[2] < 0:
            return self._bce_str % value
        else:
            return value

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "od", d1, "do", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "między", d1, "a", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("pl_PL", "polish", "Polish_Poland", "pl", ("%d.%m.%Y",)),
    DateParserPL,
    DateDisplayPL,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2012  Brian G. Matherly
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Matt Keenan (matt.keenan@gmail.com)
# Copyright (C) 2013-2014  Paul Franklin
# Copyright (C) 2010,2015  Craig J. Anderson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Reports/Text Reports/Descendant Report.
"""

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    FONT_SANS_SERIF,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
)
from gramps.gen.plug.menu import (
    NumberOption,
    PersonOption,
    BooleanOption,
    EnumeratedListOption,
)
from gramps.gen.errors import ReportError
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.db import (
    get_birth_or_fallback,
    get_death_or_fallback,
    get_marriage_or_fallback,
    get_divorce_or_fallback,
)
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.display.place import displayer as _pd
from gramps.gen.display.name import displayer as _nd


# ------------------------------------------------------------------------
#
# PrintDAboville
#
# ------------------------------------------------------------------------
class PrintDAboville:
    """
    d'Aboville numbering system

    (according to en.wikipedia.org/Genealogical_numbering_systems
    his name is spelled "d'Aboville" and not "D'Aboville" but I will
    leave this class name alone, mainly fixing the translated string,
    so that it is both accurate and also agrees with the DDR string)
    """

    def __init__(self):
        self.num = [0]

    def number(self, level):
        """Make the current number based upon the current level"""
        # Set up the array based on the current level
        while len(self.num) > level:  # We can go from a level 8 to level 2
            self.num.pop()
        if len(self.num) < level:
            self.num.append(0)

        # Increment the current level - initalized with 0
        self.num[-1] += 1

        # Display
        return ".".join(map(str, self.num))


# ------------------------------------------------------------------------
#
# PrintHenry
#
# ------------------------------------------------------------------------
class PrintHenry:
    """Henry numbering system"""

    def __init__(self, modified=False):
        self.henry = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.modified = modified
        self.num = [0]

    def number(self, level):
        """Make the current number based upon the current level"""
        # Set up the array based on the current level
        while len(self.num) > level:  # We can go from a level 8 to level 2
            self.num.pop()
        if len(self.num) < level:
            self.num.append(0)

        # Incriment the current level - initalized with 0
        self.num[-1] += 1

        def strd(inti):
            """no change needed"""
            return "(" + str(inti) + ")"

        # Display
        if self.modified is False:
            return "".join(
                map(
                    lambda x: self.henry[x - 1] if x <= len(self.henry) else strd(x),
                    self.num,
                )
            )
        else:
            return "".join(map(lambda x: str(x) if x < 10 else strd(x), self.num))


# ------------------------------------------------------------------------
#
# PrintSimple
#
# ------------------------------------------------------------------------
class PrintSimple:
    """Simple numbering system"""

    def __init__(self, showdups):
        self.showdups = showdups
        self.num = {0: 1}

    def number(self, level):
        """the number of the person"""
        if self.showdups:
            # Just show original simple numbering
            to_return = "%d." % level
        else:
            to_return = str(level)
            if level > 1:
                to_return += "-" + str(self.num[level - 1])
            to_return += "."

            self.num[level] = 1
            self.num[level - 1] = self.num[level - 1] + 1

        return to_return


# ------------------------------------------------------------------------
#
# PrintVlliers
#   de_Villiers_Pama numbering system
#
# ------------------------------------------------------------------------
class PrintVilliers:
    """de_Villiers_Pama numbering system"""

    def __init__(self):
        self.pama = "abcdefghijklmnopqrstuvwxyz"
        self.num = {0: 1}

    def number(self, level):
        """the number of the person"""
        to_return = self.pama[level - 1]
        if level > 1:
            to_return += str(self.num[level - 1])
        to_return += "."

        self.num[level] = 1
        self.num[level - 1] = self.num[level - 1] + 1

        return to_return


# ------------------------------------------------------------------------
#
# class PrintMeurgey
#   Meurgey_de_Tupigny numbering system
#
# ------------------------------------------------------------------------
class PrintMeurgey:
    """Meurgey_de_Tupigny numbering system"""

    def __init__(self):
        self.childnum = [""]

    def number(self, level):
        """the number of the person"""
        if level == 1:
            dash = ""
        else:
            dash = "-"
            if len(self.childnum) < level:
                self.childnum.append(1)

        to_return = utils.roman(level) + dash + str(self.childnum[level - 1]) + "."

        if level > 1:
            self.childnum[level - 1] += 1

        return to_return


# ------------------------------------------------------------------------
#
# Printinfo
#
# ------------------------------------------------------------------------
class Printinfo:
    """
    A base class used to help make the individual numbering system classes.
    This class must first be initialized with set_class_vars
    """

    def __init__(
        self,
        doc,
        database,
        numbering,
        showmarriage,
        showdivorce,
        showlifespan,
        name_display,
        rlocale,
        want_ids,
        pformat,
    ):
        # classes
        self._name_display = name_display
        self.doc = doc
        self.database = database
        self.numbering = numbering
        # variables
        self.showmarriage = showmarriage
        self.showdivorce = showdivorce
        self.showlifespan = showlifespan
        self.want_ids = want_ids
        self._ = rlocale.translation.sgettext  # needed for English
        self._get_date = rlocale.get_date
        self.pformat = pformat

    def __date_place(self, event):
        """return the date and/or place an event happened"""
        if event:
            date = self._get_date(event.get_date_object())
            place_handle = event.get_place_handle()
            if place_handle:
                place = _pd.display_event(self.database, event, self.pformat)
                return "%(event_abbrev)s %(date)s - %(place)s" % {
                    "event_abbrev": event.type.get_abbreviation(self._),
                    "date": date,
                    "place": place,
                }
            else:
                return "%(event_abbrev)s %(date)s" % {
                    "event_abbrev": event.type.get_abbreviation(self._),
                    "date": date,
                }
        return ""

    def dump_string(self, person, family=None):
        """generate a descriptive string for a person"""
        string = ""

        if self.showlifespan:
            string += self.__date_place(get_birth_or_fallback(self.database, person))
            tmp = self.__date_place(get_death_or_fallback(self.database, person))
            if string and tmp:
                string += self._(", ")  # Arabic OK
            string += tmp

            if string:
                string = " (" + string + ")"

        if family and self.showmarriage:
            tmp = self.__date_place(get_marriage_or_fallback(self.database, family))
            if tmp:
                string += self._(", ") + tmp  # Arabic OK

        if family and self.showdivorce:
            tmp = self.__date_place(get_divorce_or_fallback(self.database, family))
            if tmp:
                string += self._(", ") + tmp  # Arabic OK

        if family and self.want_ids:
            string += " (%s)" % family.get_gramps_id()

        self.doc.write_text(string)

    def print_person(self, level, person):
        """print the person"""
        display_num = self.numbering.number(level)
        self.doc.start_paragraph("DR-Level%d" % min(level, 32), display_num)
        mark = utils.get_person_mark(self.database, person)
        self.doc.write_text(self._name_display.display(person), mark)
        if self.want_ids:
            self.doc.write_text(" (%s)" % person.get_gramps_id())
        self.dump_string(person)
        self.doc.end_paragraph()
        return display_num

    def print_spouse(self, level, spouse_handle, family_handle):
        """print the spouse"""
        # Currently print_spouses is the same for all numbering systems.
        if spouse_handle:
            spouse = self.database.get_person_from_handle(spouse_handle)
            mark = utils.get_person_mark(self.database, spouse)
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            name = self._name_display.display(spouse)
            self.doc.write_text(self._("sp. %(spouse)s") % {"spouse": name}, mark)
            if self.want_ids:
                self.doc.write_text(" (%s)" % spouse.get_gramps_id())
            self.dump_string(spouse, family_handle)
            self.doc.end_paragraph()
        else:
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            self.doc.write_text(
                self._("sp. %(spouse)s") % {"spouse": self._("Unknown")}
            )
            self.doc.end_paragraph()

    def print_reference(self, level, person, display_num):
        """print the reference"""
        # Person and their family have already been printed so
        # print reference here
        if person:
            mark = utils.get_person_mark(self.database, person)
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            name = self._name_display.display(person)
            self.doc.write_text(
                self._("sp. see %(reference)s: %(spouse)s")
                % {"reference": display_num, "spouse": name},
                mark,
            )
            self.doc.end_paragraph()


# ------------------------------------------------------------------------
#
# RecurseDown
#
# ------------------------------------------------------------------------
class RecurseDown:
    """
    A simple object to recurse from a person down through their descendants

    The arguments are:

    max_generations: The max number of generations
    database:  The database object
    obj_print: A Printinfo derived class that prints person
               information on the report
    """

    def __init__(self, max_generations, database, obj_print, showdups, rlocale):
        self.max_generations = max_generations
        self.database = database
        self.obj_print = obj_print
        self.showdups = showdups
        self.person_printed = {}
        self._ = rlocale.translation.sgettext  # needed for English

    def recurse(self, level, person, curdepth):
        """recurse"""

        person_handle = person.get_handle()
        display_num = self.obj_print.print_person(level, person)

        if curdepth is None:
            ref_str = display_num
        else:
            ref_str = curdepth + " " + display_num

        if person_handle not in self.person_printed:
            self.person_printed[person_handle] = ref_str

        for family_handle in person.get_family_handle_list():
            family = self.database.get_family_from_handle(family_handle)

            spouse_handle = utils.find_spouse(person, family)

            if not self.showdups and spouse_handle in self.person_printed:
                # Just print a reference
                spouse = self.database.get_person_from_handle(spouse_handle)
                self.obj_print.print_reference(
                    level, spouse, self.person_printed[spouse_handle]
                )
            else:
                self.obj_print.print_spouse(level, spouse_handle, family)

                if spouse_handle:
                    spouse_num = self._("%s sp.") % ref_str
                    self.person_printed[spouse_handle] = spouse_num

                if level >= self.max_generations:
                    continue

                childlist = family.get_child_ref_list()[:]
                for child_ref in childlist:
                    child = self.database.get_person_from_handle(child_ref.ref)
                    self.recurse(level + 1, child, ref_str)


# ------------------------------------------------------------------------
#
# DescendantReport
#
# ------------------------------------------------------------------------
class DescendantReport(Report):
    """Descendant report"""

    def __init__(self, database, options, user):
        """
        Create the DescendantReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        gen           - Maximum number of generations to include.
        name_format   - Preferred format to display names
        dups          - Whether to include duplicate descendant trees
        incl_private  - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        inc_id        - Whether to include Gramps IDs
        """

        Report.__init__(self, database, options, user)

        menu = options.menu

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)

        self.max_generations = menu.get_option_by_name("gen").get_value()
        self.want_ids = menu.get_option_by_name("inc_id").get_value()

        pid = menu.get_option_by_name("pid").get_value()
        self.center_person = self.database.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        # Initialize the Printinfo class
        self._showdups = menu.get_option_by_name("dups").get_value()
        numbering = menu.get_option_by_name("numbering").get_value()
        if numbering == "Simple":
            obj = PrintSimple(self._showdups)
        elif numbering == "Henry":
            obj = PrintHenry()
        elif numbering == "Modified Henry":
            obj = PrintHenry(modified=True)
        elif numbering == "d'Aboville":
            obj = PrintDAboville()
        elif numbering == "de Villiers/Pama":
            obj = PrintVilliers()
        elif numbering == "Meurgey de Tupigny":
            obj = PrintMeurgey()
        else:
            raise AttributeError("no such numbering: '%s'" % numbering)

        marrs = menu.get_option_by_name("marrs").get_value()
        divs = menu.get_option_by_name("divs").get_value()
        lifespan = menu.get_option_by_name("lifespan").get_value()

        stdoptions.run_name_format_option(self, menu)

        pformat = menu.get_option_by_name("place_format").get_value()

        self.obj_print = Printinfo(
            self.doc,
            self.database,
            obj,
            marrs,
            divs,
            lifespan,
            self._name_display,
            self._locale,
            self.want_ids,
            pformat,
        )

    def write_report(self):
        self.doc.start_paragraph("DR-Title")
        name = self._name_display.display(self.center_person)
        # feature request 2356: avoid genitive form
        title = self._("Descendants of %s") % name
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        recurse = RecurseDown(
            self.max_generations,
            self.database,
            self.obj_print,
            self._showdups,
            self._locale,
        )
        recurse.recurse(1, self.center_person, None)


# ------------------------------------------------------------------------
#
# DescendantOptions
#
# ------------------------------------------------------------------------
class DescendantOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        category_name = _("Report Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)

        numbering = EnumeratedListOption(_("Numbering system"), "Simple")
        numbering.set_items(
            [
                ("Simple", _("Simple numbering")),
                ("d'Aboville", _("d'Aboville numbering")),
                ("Henry", _("Henry numbering")),
                ("Modified Henry", _("Modified Henry numbering")),
                ("de Villiers/Pama", _("de Villiers/Pama numbering")),
                ("Meurgey de Tupigny", _("Meurgey de Tupigny numbering")),
            ]
        )
        numbering.set_help(_("The numbering system to be used"))
        menu.add_option(category_name, "numbering", numbering)

        gen = NumberOption(_("Generations"), 10, 1, 15)
        gen.set_help(_("The number of generations to include in the report"))
        menu.add_option(category_name, "gen", gen)

        stdoptions.add_gramps_id_option(menu, category_name)

        marrs = BooleanOption(_("Show marriage info"), False)
        marrs.set_help(_("Whether to show marriage information in the report."))
        menu.add_option(category_name, "marrs", marrs)

        divs = BooleanOption(_("Show divorce info"), False)
        divs.set_help(_("Whether to show divorce information in the report."))
        menu.add_option(category_name, "divs", divs)

        lifespan = BooleanOption(_("Show birth and death info"), True)
        lifespan.set_help(
            _("Whether to show birth and death information in the report.")
        )
        menu.add_option(category_name, "lifespan", lifespan)

        dups = BooleanOption(_("Show duplicate trees"), True)
        dups.set_help(_("Whether to show duplicate Family Trees in the report."))
        menu.add_option(category_name, "dups", dups)

        category_name = _("Report Options (2)")

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_place_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

    def make_default_style(self, default_style):
        """Make the default output style for the Descendant Report."""
        fstyle = FontStyle()
        fstyle.set_size(12)
        fstyle.set_type_face(FONT_SANS_SERIF)
        fstyle.set_bold(1)
        pstyle = ParagraphStyle()
        pstyle.set_header_level(1)
        pstyle.set_bottom_border(1)
        pstyle.set_top_margin(utils.pt2cm(3))
        pstyle.set_bottom_margin(utils.pt2cm(3))
        pstyle.set_font(fstyle)
        pstyle.set_alignment(PARA_ALIGN_CENTER)
        pstyle.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("DR-Title", pstyle)

        fstyle = FontStyle()
        fstyle.set_size(10)
        for i in range(1, 33):
            pstyle = ParagraphStyle()
            pstyle.set_font(fstyle)
            pstyle.set_top_margin(utils.pt2cm(fstyle.get_size() * 0.125))
            pstyle.set_bottom_margin(utils.pt2cm(fstyle.get_size() * 0.125))
            pstyle.set_first_indent(-0.5)
            pstyle.set_left_margin(min(10.0, float(i - 0.5)))
            pstyle.set_description(_("The style used for the level %d display.") % i)
            default_style.add_paragraph_style("DR-Level%d" % min(i, 32), pstyle)

            pstyle = ParagraphStyle()
            pstyle.set_font(fstyle)
            pstyle.set_top_margin(utils.pt2cm(fstyle.get_size() * 0.125))
            pstyle.set_bottom_margin(utils.pt2cm(fstyle.get_size() * 0.125))
            pstyle.set_left_margin(min(10.0, float(i - 0.5)))
            pstyle.set_description(
                _("The style used for the spouse level %d display.") % i
            )
            default_style.add_paragraph_style("DR-Spouse%d" % min(i, 32), pstyle)
